
from bokeh.models import ColumnDataSource, Slider, CustomJS, BoxZoomTool, Toggle
from bokeh.models.glyphs import Patch, Patches
from bokeh.models.ranges import Range1d
from bokeh.layouts import row, column
from bokeh.plotting import figure, show
from tramway.tessellation.base import Partition
from .map import *
import numpy as np
import itertools
import warnings
import polytope as pt


class RoiController(object):
    def __init__(self, roi_collection=None,
                patches_x=None, patches_y=None,
                xlim=None, ylim=None,
                properties=('line_color','line_width'),
                default_values=('black',1),
                active_values=('green',2),
                **patches_kwargs):
        self.active_roi = None
        self.dynamic_properties = properties
        self.property_default_values = default_values
        self.property_active_values = active_values
        if patches_x is None or patches_y is None:
            patches_x, patches_y = [], []
            for bounding_box in roi_collection.bounding_box():
                patch_x = bounding_box[[0,0,2,2]]
                patch_y = bounding_box[[1,3,3,1]]
                patches_x.append(patch_x)
                patches_y.append(patch_y)
        self._patches_kwargs = patches_kwargs
        patch_data = dict(xs=patches_x, ys=patches_y)
        patch_data.update({prop: [value for _ in patches_x]
                for prop, value in zip(self.dynamic_properties, self.property_default_values)})
        patch_data['active'] = [0 for _ in patches_x]
        self.patches = ColumnDataSource(patch_data)
        self.line = ColumnDataSource(dict(x=[0,0,np.nan,0,0,np.nan,0,0,np.nan,0,0], y=[0,0,np.nan,0,0,np.nan,0,0,np.nan,0,0]))
        self._xlim = xlim
        self._ylim = ylim
    @property
    def patches(self):
        return self._patches
    @patches.setter
    def patches(self, ps):
        self._patches = ps
        self._make_prop_setter()
    @property
    def independent_patches(self):
        return isinstance(self.patches, (list, tuple))
    def _make_prop_setter(self):
        if self.independent_patches:
            if not self.patches:
                raise ValueError('empty sequence of patches')
            sample_patch = patch[0]
            if isinstance(sample_patch, ColumnDataSource):
                def _set(patch, prop, val):
                    assert patch != 'all'
                    patch.data[prop] = val
            elif isinstance(sample_patch, Patch):
                def _set(patch, prop, val):
                    assert patch != 'all'
                    setattr(patch.glyph, prop, val)
            else:
                raise TypeError('wrong patch type')
        else:
            if isinstance(self.patches, ColumnDataSource):
                def _all_patches_get_prop(prop):
                    return self.patches.data[prop]
                def _all_patches_set_prop(prop, values):
                    self.patches.data[prop] = values
            elif isinstance(self.patches, Patches):
                def _all_patches_get_prop(prop):
                    return getattr(self.patches, prop)
                def _all_patches_set_prop(prop, values):
                    setattr(self.patches, prop, values)
            else:
                raise TypeError('wrong multi-patch type')
            def _set(patch_index, prop, val):
                _all_patches_prop = _all_patches_get_prop(prop)
                if patch_index == 'all':
                    _all_patches_set_prop(prop, [val for _ in _all_patches_prop])
                else:
                    _all_patches_prop[patch_index] = val
        self._set = _set
    def _set_props(self, patch, property_values):
        if self.independent_patches:
            if patch == 'all':
                for patch in self.patches:
                    for prop, val in zip(self.dynamic_properties, property_values):
                        self._set(patch, prop, val)
                return # stop here
            else:
                patch = self.patches[patch]
        for prop, val in zip(self.dynamic_properties, property_values):
            self._set(patch, prop, val)
    def set_active(self, new_active_roi):
        if new_active_roi != self.active_roi and self.active_roi is not None:
            self.unset_active(self.active_roi)
        self._set_props(new_active_roi, self.property_active_values)
        #
        x0, x3 = self.xlim
        y0, y3 = self.ylim
        xs, ys = self.patches.data['xs'][new_active_roi], self.patches.data['ys'][new_active_roi]
        x1, x2 = xs[0], xs[2]
        y1, y2 = ys[0], ys[1]
        xc, yc = .5*(x1+x2), .5*(y1+y2)
        self.line.data['x'] = [x0, x1, np.nan, x2, x3, np.nan, xc, xc, np.nan, xc, xc]
        self.line.data['y'] = [yc, yc, np.nan, yc, yc, np.nan, y0, y1, np.nan, y2, y3]
        #
        self.active_roi = new_active_roi
    def unset_active(self, roi):
        self._set_props(roi, self.property_default_values)
        if roi == self.active_roi or roi == 'all':
            self.active_roi = None
    def js_callback(self, roi_plot):
        default_props = dict(zip(self.dynamic_properties, self.property_default_values))
        active_props = dict(zip(self.dynamic_properties, self.property_active_values))
        return CustomJS(args=dict(patches=self.patches, lines=self.line, roi_plot=roi_plot), code="""
                var active = patches.data['active'][0];
                var line_color = patches.data['line_color'];
                var line_width = patches.data['line_width'];
                line_color[active] = '{}';
                line_width[active] = {};
                var new_active = this.value-1;
                line_color[new_active] = '{}';
                line_width[new_active] = {};
                patches.data['active'][0] = new_active;
                var x = patches.data['xs'][new_active];
                var xmin = x[0], xmax = x[2];
                var xc = 0.5 * (xmin + xmax);
                var y = patches.data['ys'][new_active];
                var ymin = y[0], ymax = y[1];
                var yc = 0.5 * (ymin + ymax);
                x = lines.data['x'];
                x[1] = xmin, x[3] = xmax;
                for (var i of [6,7,9,10]) {{
                    x[i] = xc;
                }}
                y = lines.data['y'];
                for (i of [0,1,3,4]) {{
                    y[i] = yc;
                }}
                y[7] = ymin, y[9] = ymax;
                var width = xmax-xmin, height = ymax-ymin;
                if (width < height) {{
                    xmin = xc - 0.5 * height;
                    xmax = xc + 0.5 * height;
                }} else if (height < width) {{
                    ymin = yc - 0.5 * width;
                    ymax = yc + 0.5 * width;
                }}
                roi_plot.x_range.start = xmin;
                roi_plot.x_range.end   = xmax;
                roi_plot.y_range.start = ymin;
                roi_plot.y_range.end   = ymax;
                lines.change.emit();
                patches.change.emit();
                roi_plot.change.emit();
            """.format(
                default_props['line_color'],
                default_props['line_width'],
                active_props['line_color'],
                active_props['line_width']))
    @property
    def patches_kwargs(self):
        return dict(xs='xs', ys='ys', line_color='line_color', line_width='line_width',
                    source=self.patches, **self._patches_kwargs)
    @property
    def line_kwargs(self):
        active_props = dict(zip(self.dynamic_properties, self.property_active_values))
        return dict(x='x', y='y', source=self.line, **active_props)
    @property
    def xlim(self):
        if self._xlim is None:
            xs = list(itertools.chain(*self.patches.data['xs']))
            self._xlim = min(xs), max(xs)
        return self._xlim
    @xlim.setter
    def xlim(self, xl):
        self._xlim = xl
    @property
    def ylim(self):
        if self._ylim is None:
            ys = list(itertools.chain(*self.patches.data['ys']))
            self._ylim = min(ys), max(ys)
        return self._ylim
    @ylim.setter
    def ylim(self, yl):
        self._ylim = yl

def match_aspect(xlim, ylim):
    width, height = xlim[1]-xlim[0], ylim[1]-ylim[0]
    if width < height:
        xc = .5*(xlim[0]+xlim[1])
        xlim = [ xc-.5*height, xc+.5*height ]
    elif height < width:
        yc = .5*(ylim[0]+ylim[1])
        ylim = [ yc-.5*width, yc+.5*width ]
    return xlim, ylim


class RoiCollection(object):
    def __init__(self, global_tessellation=None, connected_components=None,
            roi_tessellations=None, roi_meta_tessellation=None,
            bounding_boxes=None):
        self.global_tessellation = global_tessellation
        self.roi_tessellations = roi_tessellations
        self.roi_meta_tessellation = roi_meta_tessellation
        self.connected_components = connected_components
        self._bounding_boxes = bounding_boxes
    def bounding_box(self, margin=None):
        if self._bounding_boxes is None:
            bb = []
            if self.roi_meta_tessellation is not None:
                meta_mesh = self.roi_meta_tessellation
                for vertex_indices in meta_mesh.cell_vertices:
                    lower_bound = meta_mesh.vertices[vertex_indices[0]]
                    upper_bound = meta_mesh.vertices[vertex_indices[2]]
                    bounding_box = np.r_[lower_bound, upper_bound]
                    bb.append(bounding_box)
            elif self.roi_tessellations is not None:
                for roi in self.roi_tessellations:
                    lower_bound = roi.bounding_box[['x','y']]['min']
                    upper_bound = roi.bounding_box[['x','y']]['max']
                    bounding_box = np.r_[lower_bound, upper_bound]
                    bb.append(bounding_box)
            else:
                tessellation = self.global_tessellation
                for component in self.connected_components:
                    ## center
                    #centers = tessellation.cell_centers[component]
                    #center = np.mean(centers, axis=0)
                    # bounding box
                    vertex_indices = set()
                    for cell in component:
                        vertex_indices |= set(tessellation.cell_vertices[cell])
                    try:
                        vertex_indices.remove(-1)
                    except KeyError:
                        pass
                    vertices = tessellation.vertices[list(vertex_indices)]
                    lower_bound, upper_bound = np.min(vertices, axis=0), np.max(vertices, axis=0)
                    bounding_box = np.r_[lower_bound, upper_bound]
                    #
                    #bb.append((center, bounding_box))
                    bb.append(bounding_box)
            ## sanity check
            #polygons = [ pt.box2poly([[b[0],b[2]],[b[1],b[3]]]) for b in bb ]
            #if not all([ all([ pt.is_empty(p.intersect(q)) for q in polygons[i+1:] ]) for i, p in enumerate(polygons) ]):
            #    warnings.warn('some rois overlap', RuntimeWarning)
            #
            self._bounding_boxes = bb
        if margin:
            bb = []
            for bounding_box in self._bounding_boxes:
                lower_bound = bounding_box[[0,1]]
                upper_bound = bounding_box[[2,3]]
                bounding_box = np.r_[lower_bound - margin, upper_bound + margin]
                bb.append(bounding_box)
        else:
            bb = self._bounding_boxes
        return bb
    def __len__(self):
        if self._bounding_boxes is None:
            return len(self.connected_components)
        else:
            return len(self._bounding_boxes)


class RoiBrowser(object):
    def __init__(self, global_map_cells=None, global_map_values=None, # these two args first for bw comp
            roi_tessellations=None, roi_map_values=None, roi_meta_tessellation=None,
            **kwargs):
        self.global_map_cells  = global_map_cells
        self.global_map_values = global_map_values
        self.points = None if roi_meta_tessellation is None else roi_meta_tessellation.points
        tessellation = lambda a: None if a is None else a.tessellation
        self.roi_model = RoiCollection(
                global_tessellation=tessellation(global_map_cells),
                roi_tessellations=roi_tessellations,
                roi_meta_tessellation=tessellation(roi_meta_tessellation),
                **kwargs)
        self.roi_map_values = roi_map_values
        self.first_active_roi = 0

        # options for view elements
        self.full_fov_figure_kwargs = dict(toolbar_location='above', toolbar_sticky=False, match_aspect=True, tools='pan, wheel_zoom, reset')
        self.zooming_in_figure_kwargs = dict(toolbar_location=None, active_drag=None, match_aspect=True, tools='pan, wheel_zoom, reset')
        self.scalar_map_2d_kwargs = {}
        self.points_kwargs = dict(color='r', alpha=.1)
        self.trajectories_kwargs = dict(color='r', line_width=.5, line_alpha=.5, loc_alpha=.1, loc_size=6)
        self.slider_kwargs = dict(title='roi')

        self.roi_controller = RoiController(self.roi_model, fill_color=None)
        x, y = self.points['x'].values, self.points['y'].values
        self.roi_controller.xlim = [x.min(), x.max()]
        self.roi_controller.ylim = [y.min(), y.max()]

    @property
    def values(self): # for bw compatibility during code transition
        warnings.warn('deprecated attribute: values', RuntimeWarning)
        return self.global_map_cells

    @property
    def points(self):
        if self._points is None:
            if self.global_map_cells is not None:
                self._points = self.global_map_cells.points
        return self._points

    @points.setter
    def points(self, pts):
        self._points = pts

    @property
    def cells(self):
        warnings.warn('deprecated attribute: cells', RuntimeWarning)
        return self.global_map_cells

    @property
    def roi_tessellations(self):
        return self.roi_model.roi_tessellations

    def full_fov(self):
        ctrl = self.roi_controller
        full_fov_fig = figure(**self.full_fov_figure_kwargs)
        full_fov_fig.add_tools(BoxZoomTool(match_aspect=True))
        if self.global_map_values is not None:
            scalar_map_2d(self.global_map_cells, self.global_map_values,
                    figure=full_fov_fig, **self.scalar_map_2d_kwargs)
        elif False:
            _min, _max = zip(*[ (values.min(), values.max()) for values in self.roi_map_values ])
            clim = (min(_min), max(_max))
            for cells,values in zip(self.roi_tessellations,self.roi_map_values):
                scalar_map_2d(cells, values, figure=full_fov_fig, clim=clim,
                        **self.scalar_map_2d_kwargs)
        plot_points(self.points, figure=full_fov_fig, **self.points_kwargs)
        full_fov_fig.patches(**ctrl.patches_kwargs)
        full_fov_fig.line(**ctrl.line_kwargs)
        ctrl.unset_active('all')
        ctrl.set_active(self.first_active_roi)
        self.roi_view_full_fov = full_fov_fig
        return full_fov_fig

    def zooming_in(self):
        ctrl = self.roi_controller
        roi_bb = self.roi_model.bounding_box(margin=.1)[self.first_active_roi]
        xlim = roi_bb[[0,2]].tolist()
        ylim = roi_bb[[1,3]].tolist()
        xlim, ylim = match_aspect(xlim, ylim)
        zooming_in_fig = figure(**self.zooming_in_figure_kwargs)
        zooming_in_fig.add_tools(BoxZoomTool(match_aspect=True))
        if self.global_map_values is not None:
            scalar_map_2d(self.global_map_cells, self.global_map_values, figure=zooming_in_fig,
                    **self.scalar_map_2d_kwargs)
        else:
            for cells,values in zip(self.roi_tessellations,self.roi_map_values):
                scalar_map_2d(cells, values, figure=zooming_in_fig,
                        **self.scalar_map_2d_kwargs)
        traj_handles = plot_trajectories(self.points, figure=zooming_in_fig, **self.trajectories_kwargs)
        self.trajectory_handles = traj_handles[0::2]
        self.location_handles = traj_handles[1::2]
        zooming_in_fig.x_range = Range1d(*xlim)
        zooming_in_fig.y_range = Range1d(*ylim)
        self.roi_view_zooming_in = zooming_in_fig
        return zooming_in_fig

    def slider(self):
        ctrl = self.roi_controller
        zooming_in_fig = self.roi_view_zooming_in
        slider = Slider(start=1, end=len(self.roi_model), step=1, value=self.first_active_roi+1, **self.slider_kwargs)
        slider.js_on_change('value_throttled', ctrl.js_callback(zooming_in_fig))
        self.roi_view_slider = slider
        return slider

    def visibility_button1(self):
        self.trajectory_visibility_button = Toggle(label='Hide lines', button_type='success')
        assert not self.trajectory_handles[1:]
        def set_visibility(multiline):
            return CustomJS(args=dict(multiline=multiline), code="""
                    multiline.visible=!multiline.visible;
                    if (multiline.visible) {
                        this.label='Hide lines';
                    } else {
                        this.label='Show lines';
                    }
                    """)
        self.trajectory_visibility_button.js_on_click(set_visibility(self.trajectory_handles[0]))
        return self.trajectory_visibility_button

    def visibility_button2(self):
        self.location_visibility_button = Toggle(label='Hide points', button_type='success')
        assert not self.location_handles[1:]
        def set_visibility(multiline):
            return CustomJS(args=dict(multiline=multiline), code="""
                    multiline.visible=!multiline.visible;
                    if (multiline.visible) {
                        this.label='Hide points';
                    } else {
                        this.label='Show points';
                    }
                    """)
        self.location_visibility_button.js_on_click(set_visibility(self.location_handles[0]))
        return self.location_visibility_button

    def make_default_view(self):
        full_fov_map = self.full_fov()
        zooming_in_map = self.zooming_in()
        visibility_button1 = self.visibility_button1()
        visibility_button2 = self.visibility_button2()
        if 1 < len(self.roi_model):
            slider = self.slider()
            self.roi_view = row(zooming_in_map, column(slider, full_fov_map, row(visibility_button1, visibility_button2), sizing_mode='scale_width'))
        else:
            self.roi_view = row(zooming_in_map, column(full_fov_map, sizing_mode='scale_width'))

    def show(self):
        show(self.roi_view)


def roi_plot(roi, *args, **kwargs):
    from tramway.helper import map_plot
    from .map import scalar_map_2d, plot_points

    feature_name = kwargs.pop('feature')

    kwargs.update(dict(use_bokeh=True, show=False))
    if 'point_style' not in kwargs:
        kwargs['point_style'] = dict(color='r', alpha=.1)
    fig = map_plot(*args, **kwargs)
    full_fov_fig = fig[0]

    patches_x, patches_y = [], []
    for center, bounding_box in roi:
        patch_x = bounding_box[[0,0,2,2]]
        patch_y = bounding_box[[1,3,3,1]]
        patches_x.append(patch_x)
        patches_y.append(patch_y)

    roi_controller = RoiController(patches_x, patches_y, fill_color=None)
    full_fov_fig.patches(**roi_controller.patches_kwargs)

    roi_controller.unset_active('all')
    first_roi = 0
    roi_controller.set_active(first_roi)

    # right panel

    first_roi_bb = roi[first_roi][1]
    xlim = first_roi_bb[[0,2]].tolist()
    ylim = first_roi_bb[[1,3]].tolist()

    from bokeh.plotting import figure, show

    analysis_tree = args[0]
    partition_label, map_label = kwargs['label']
    cells = analysis_tree[partition_label].data
    _map = analysis_tree[partition_label][map_label].data[feature_name]

    _kwargs = {}
    for _attr in ('clim',):
        if _attr in kwargs:
            _kwargs[attr] = kwargs[attr]

    focused_fig = figure()
    scalar_map_2d(cells, _map, figure=focused_fig, xlim=xlim, ylim=xlim, **_kwargs)
    plot_points(cells, color='r', alpha=.1)

    from bokeh.models import Slider, CustomJS
    slider = Slider(start=1, end=len(roi), step=1, value=1, title='roi')
    slider.js_on_change('value', roi_controller.js_callback(focused_fig))

    # plot with layout

    from bokeh.layouts import row, column
    show(row(full_fov_fig, column(focused_fig, slider, sizing_mode='scale_width')))




# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import math
import numpy as np
import pandas as pd
from numpy.polynomial import polynomial as poly
from collections import OrderedDict


def delta0(cells, i, X, index_map=None, **kwargs):
    r"""
    Differences with neighbour values:

    .. math::

        \Delta X_i = \frac{1}{\sqrt{|\mathcal{N}_i|}} \left[ \frac{X_i-X_j}{|| \textbf{x}_i-\textbf{x}_j ||} \right]_{j \in \mathcal{N}_i}

    The above scaling is chosen so that combining the element-wise square of :meth:`~tramway.inference.base.Distributed.local_variation` with :meth:`~tramway.inference.base.Distributed.grad_sum` results in the following scalar penalty:

    .. math::

        \Delta X_i^2 = \frac{1}{ | \mathcal{N}_i | } \sum_{j \in \mathcal{N}_i} \left( \frac{X_i-X_j}{|| \textbf{x}_i-\textbf{x}_j ||} \right)^2

    Claims cache variable '*delta0*'.

    Arguments:

        cells (tramway.inference.base.Distributed):
            distributed cells.

        i (int):
            cell index at which the differences are evaluated.

        X (array):
            vector of a scalar measurement at every cell.

        index_map (array):
            index map that converts cell indices to indices in X.

    Returns:

        array:
            difference vector with as many elements as there are neighbours.

    """
    cell = cells[i]
    # below, the measurement is renamed y and the coordinates are X
    y = X

    # cache neighbours (indices and center locations)
    if not isinstance(cell.cache, dict):
        cell.cache = {}
    try:
        i, adjacent, dx_norm = cell.cache['delta0']
    except KeyError:
        adjacent = _adjacent = cells.neighbours(i)
        if index_map is not None:
            adjacent = index_map[_adjacent]
            ok = 0 <= adjacent
            if not np.all(ok):
                adjacent, _adjacent = adjacent[ok], _adjacent[ok]
        if _adjacent.size:
            x0 = cell.center[np.newaxis,:]
            x = np.vstack([ cells[j].center for j in _adjacent ])
            dx_norm = x - x0
            dx_norm = np.sqrt(np.sum(dx_norm * dx_norm, axis=1))
        else:
            dx_norm = None

        if index_map is not None:
            i = index_map[i]
        cell.cache['delta0'] = (i, adjacent, dx_norm)

    if dx_norm is None:
        return None

    y0, y = y[i], y[adjacent]

    # scale by the number of differences to make the sum of the returned values be a mean value instead
    return (y - y0) / dx_norm / np.sqrt(float(len(y)))


def delta0_without_scaling(cells, i, X, index_map=None, **kwargs):
    cell = cells[i]
    # below, the measurement is renamed y and the coordinates are X
    y = X

    # cache neighbours (indices and center locations)
    if not isinstance(cell.cache, dict):
        cell.cache = {}
    try:
        i, adjacent, dx_norm = cell.cache['delta0']
    except KeyError:
        adjacent = _adjacent = cells.neighbours(i)
        if index_map is not None:
            adjacent = index_map[_adjacent]
            ok = 0 <= adjacent
            if not np.all(ok):
                adjacent, _adjacent = adjacent[ok], _adjacent[ok]
        if _adjacent.size:
            x0 = cell.center[np.newaxis,:]
            x = np.vstack([ cells[j].center for j in _adjacent ])
            dx_norm = x - x0
            dx_norm = np.sqrt(np.sum(dx_norm * dx_norm, axis=1))
        else:
            dx_norm = None

        if index_map is not None:
            i = index_map[i]
        cell.cache['delta0'] = (i, adjacent, dx_norm)

    if dx_norm is None:
        return None

    y0, y = y[i], y[adjacent]

    # scale by the number of differences to make the sum of the returned values be a mean value instead
    return (y - y0) / dx_norm


def gradn(cells, i, X, index_map=None):
    """
    Local gradient by multi-dimensional 2 degree polynomial interpolation.

    Arguments:

        cells (tramway.inference.base.Distributed):
            distributed cells.

        i (int):
            cell index at which the gradient is evaluated.

        X (array):
            vector of a scalar measurement at every cell.

        index_map (array):
            index map that converts cell indices to indices in X.

    Returns:

        array:
            local gradient vector with as many elements as there are dimensions
            in the (trans-)location data.

    """
    cell = cells[i]

    # below, the measurement is renamed Y and the coordinates are X
    Y = X
    X0 = cell.center

    # cache the components related to X
    if not isinstance(cell.cache, dict):
        cell.cache = {}
    try:
        poly = cell.cache['poly']

    except KeyError: # fill in the cache
        # find neighbours
        adjacent = _adjacent = cells.neighbours(i)
        if index_map is not None:
            i = index_map[i]
            adjacent = index_map[_adjacent]
            ok = 0 <= adjacent
            if not np.all(ok):
                adjacent, _adjacent = adjacent[ok], _adjacent[ok]
        if adjacent.size:
            X0_dup = _adjacent.size
            local = np.r_[np.full(X0_dup, i), adjacent]
            # neighbour locations
            X = np.vstack([ X0 ] * X0_dup + [ cells[j].center for j in _adjacent ])
            # cache X-related terms and indices of the neighbours
            poly = cell.cache['poly'] = (_poly2(X), local)
        else:
            poly = cell.cache['poly'] = None

    if poly is None:
        return

    V, local = poly
    Y = Y[local]

    # interpolate
    W, _, _, _ = np.linalg.lstsq(V, Y, rcond=None)
    gradX = _poly2_deriv_eval(W, X0)

    return gradX


def _poly2(X):
    n, d = X.shape
    x = X.T
    return np.hstack((
        np.vstack([ x[u] * x[v] for v in range(d) for u in range(d) ]).T,
        X,
        np.ones((n, 1), dtype=X.dtype),
        ))

def _memoize1(f):
    results = {}
    def helper(x):
        try:
            y = results[x]
        except KeyError:
            y = results[x] = f(x)
        return y
    return helper

@_memoize1
def _poly2_meta_deriv(dim):
    poly = []
    for d in range(dim):
        p = []
        for v in range(dim):
            for u in range(dim):
                if u == d:
                    if v == d:
                        p.append((2., u))
                    else:
                        p.append((1., v))
                else:
                    if v == d:
                        p.append((1., u))
                    else:
                        p.append((0., None))
        for u in range(dim):
            if u == d:
                p.append((1., None))
            else:
                p.append((0., None))
        p.append((0., None))
        poly.append(p)
    return poly

def _poly2_deriv_eval(W, X):
    if not X.shape[1:]:
        X = X[np.newaxis, :]
    dim = X.shape[1]
    dP = _poly2_meta_deriv(dim)
    Q = []
    for Pd in dP:
        Qd = 0.
        i = 0
        for q, u in Pd:
            if q == 0:
                continue
            if u is None:
                Qd += q * W[i]
            else:
                Qd += q * W[i] * X[:,u]
            i += 1
        Q.append(Qd)
    return np.hstack(Q)


default_selection_angle = .9

def get_grad_kwargs(_kwargs=None, gradient=None, grad_epsilon=None, grad_selection_angle=None, compatibility=None, grad=None, epsilon=None, **kwargs):
    """Parse :func:`grad1` keyworded arguments.

    Arguments:

        _kwargs (dict):
            mutable keyword arguments; all the keywords below are popped out.

        gradient (str): either *grad1* or *gradn*.

        grad_epsilon (float): `eps` argument for :func:`grad1` (or :func:`neighbours_per_axis`).

        grad_selection_angle (float): `selection_angle` argument for :func:`grad1` (or :func:`neighbours_per_axis`).

        compatibility (bool): backward compatibility with InferenceMAP.

        grad (str): alias for `gradient`.

        epsilon (float): alias for `grad_epsilon`.

    Returns:

        dict: keyworded arguments to :meth:`~tramway.inference.base.Distributed.grad`.

    Note: `grad` and `gradient` are currently ignored.
    """
    if _kwargs is not None:
        assert isinstance(_kwargs, dict)
        try:
            _epsilon = _kwargs.pop('epsilon')
        except KeyError:
            pass
        else:
            if epsilon is None:
                epsilon = _epsilon
            elif epsilon != _epsilon:
                # TODO: warn or raise an exception
                pass
        try:
            _grad_epsilon = _kwargs.pop('grad_epsilon')
        except KeyError:
            pass
        else:
            if grad_epsilon is None:
                grad_epsilon = _grad_epsilon
            elif grad_epsilon != _grad_epsilon:
                # TODO: warn or raise an exception
                pass
        try:
            _compatibility = _kwargs.pop('compatibility')
        except KeyError:
            pass
        else:
            if compatibility is None:
                compatibility = _compatibility
            elif compatibility != _compatibility:
                # TODO: warn or raise an exception
                pass
        try:
            _grad_selection_angle = _kwargs.pop('grad_selection_angle')
        except KeyError:
            pass
        else:
            if grad_selection_angle is None:
                grad_selection_angle = _grad_selection_angle
            elif grad_selection_angle != _grad_selection_angle:
                # TODO: warn or raise an exception
                pass

    grad_kwargs = {}
    if epsilon is not None:
        if grad_epsilon is None:
            grad_epsilon = epsilon
        elif epsilon != grad_epsilon:
            raise ValueError('`epsilon` is an alias for `grad_epsilon`; these arguments do not admit distinct values')
    if grad_epsilon is not None:
        if compatibility:
            warn('grad_epsilon breaks backward compatibility with InferenceMAP', RuntimeWarning)
        if grad_selection_angle:
            warn('grad_selection_angle is not compatible with epsilon and will be ignored', RuntimeWarning)
        grad_kwargs['eps'] = grad_epsilon
    else:
        if grad_selection_angle:
            if compatibility:
                warn('grad_selection_angle breaks backward compatibility with InferenceMAP', RuntimeWarning)
        else:
            grad_selection_angle = default_selection_angle
        if grad_selection_angle:
            grad_kwargs['selection_angle'] = grad_selection_angle
    return grad_kwargs


def neighbours_per_axis(i, cells, centers=None, eps=None, selection_angle=None):
    """
    See also :func:`grad1`.
    """
    try:
        A = cells.spatial_adjacency
    except AttributeError:
        A = cells.adjacency

    if centers is None:
        centers = [ cells[j].center for j in cells.neighbours(i) ]
        if centers:
            centers = np.vstack(centers)
        else:
            no_centers = np.zeros((cells.dim, 0), dtype=bool)
            return no_centers, no_centers

    cell = cells[i]
    center = cell.center
    below = np.zeros((centers.shape[1], centers.shape[0]), dtype=bool)
    above = np.zeros_like(below)

    if eps is None:
        if selection_angle is None or (selection_angle == .5 and centers.shape[1] == 2):
            # InferenceMAP-compatible gradient

            if A.dtype == bool:
                # divide the space in non-overlapping region and
                # assign each neighbour to a single region;
                # assign each pair of symmetric regions to a single
                # dimension (gradient calculation)
                proj = centers - center[np.newaxis, :]
                assigned_dim = np.argmax(np.abs(proj), axis=1)
                proj_dist = proj[ np.arange(proj.shape[0]), assigned_dim ]
                for j in range(cell.dim):
                    below[ j, (assigned_dim == j) & (proj_dist < 0) ] = True
                    above[ j, (assigned_dim == j) & (0 < proj_dist) ] = True

            elif A.dtype == int:
                # neighbour cells are already classified
                # as left (-1), right (1), top (2) or bottom (-2), etc
                # in the adjacency matrix
                code = A.data[A.indptr[i]:A.indptr[i+1]] - 1
                for j in range(cell.dim):
                    below[ j, code == -j ] = True
                    above[ j, code ==  j ] = True

            else:
                raise TypeError('{} adjacency labels are not supported'.format(A.dtype))

        else:
            proj = centers - center[np.newaxis, :]
            proj /= np.sqrt(np.sum(proj * proj, axis=1, keepdims=True))
            angle = 1. - 2. * np.arccos(proj) / math.pi
            for j in range(cell.dim):
                below[ j, angle[:,j] < selection_angle - 1. ] = True
                above[ j, 1. - selection_angle < angle[:,j] ] = True

    else:
        # along each dimension, divide the space in half-spaces
        # minus margin `eps` (supposed to be positive)
        for j in range(cell.dim):
            x, x0 = centers[:,j], center[j]
            below[ j, x < x0 - eps ] = True
            above[ j, x0 + eps < x ] = True

    return below, above


def grad1(cells, i, X, index_map=None, eps=None, selection_angle=None, na=np.nan):
    r"""
    Local gradient by 2 degree polynomial interpolation along each dimension independently.

    Considering spatial coordinate :math:`x`, bin :math:`i` and its neighbour bins :math:`\\mathcal{N}_i`:

    .. math::

        \left.X'_i\right|_x = \left\{
            \begin{array}{ll}
                \frac{X_i - \overline{X}_{\mathcal{N}_i}}{x_i - \overline{x}_{\mathcal{N}_i}} &
                    \textrm{ if either } \mathcal{N}_i^- \textrm{ or } \mathcal{N}_i^+ \textrm{ is } \emptyset \\
                b + 2 c x_{i} & \textrm{ with }
                    \left[ \begin{array}{ccc}
                        1 & \overline{x}_{\mathcal{N}_i^-} & \overline{x}_{\mathcal{N}_i^-}^2 \\
                        1 & x_i & x_i^2 \\
                        1 & \overline{x}_{\mathcal{N}_i^+} & \overline{x}_{\mathcal{N}_i^+}^2
                    \end{array} \right] . \left[ \begin{array}{c} a \\ b \\ c \end{array} \right] = \left[ \begin{array}{c}\overline{X}_{\mathcal{N}_i^-} \\ X_i \\ \overline{X}_{\mathcal{N}_i^+}\end{array} \right] \textrm{ otherwise } \\
            \end{array}
        \right.

    Claims cache variable *grad1*.

    Supports the InferenceMAP way of dividing the space in non-overlapping regions (quadrants in 2D)
    so that each neighbour is involved in the calculation of a single component of the gradient.

    This is the default behaviour in cases where the `adjacent` attribute contains adequate
    labels, i.e. ``-j`` and ``j`` for each dimension ``j``, representing one side and the other
    side of cell-i's center.
    This is also the default behaviour otherwise, when `eps` is ``None``.

    If `eps` is defined, the calculation of a gradient component may recruit all the points
    minus those at a projected distance smaller than this value.

    If `selection_angle` is defined, neighbours are selected in two symmetric hypercones which
    top angle is `selection_angle` times :math:`\pi` radians.

    If :func:`get_grad_kwargs` is called (most if not all the inference modes call this function),
    the default selection behaviour is equal to `selection_angle=0.9`.
    Otherwise, :func:`grad1` defaults to `selection_angle=0.5` in 2D.

    See also:

        `neighbours_per_axis`.

    Arguments:

        cells (tramway.inference.base.Distributed):
            distributed cells.

        i (int):
            cell index at which the gradient is evaluated.

        X (array):
            vector of a scalar measurement at every cell.

        index_map (array):
            index map that converts cell indices to indices in X;
            for a given cell, should be constant.

        eps (float):
            minimum projected distance from cell-i's center for neighbours to be considered
            in the calculation of the gradient;
            if `eps` is ``None``, the space is instead divided in twice as many
            non-overlapping regions as dimensions, and each neighbour is assigned to
            a single region (counts against a single dimension).
            Incompatible with `selection_angle`.

        selection_angle (float):
            top angle of the neighbour selection hypercones;
            should be in the :math:`[0.5, 1.0[` range.
            Incompatible with `eps`.

        na (float):
            value for undefined components (no neighbours).

    Returns:

        array:
            local gradient vector with as many elements as there are dimensions
            in the (trans-)location data.
    """
    assert 0 < X.size
    cell = cells[i]
    # below, the measurement is renamed y and the coordinates are X
    y = X
    X0 = cell.center

    # cache neighbours (indices and center locations)
    if not isinstance(cell.cache, dict):
        cell.cache = {}
    try:
        i, adjacent, X = cell.cache['grad1']
    except KeyError:
        adjacent = _adjacent = cells.neighbours(i)
        if index_map is not None:
            adjacent = index_map[_adjacent]
            ok = 0 <= adjacent
            if not np.all(ok):
                adjacent, _adjacent = adjacent[ok], _adjacent[ok]
        if _adjacent.size:
            X = np.vstack([ cells[j].center for j in _adjacent ])
            below, above = neighbours_per_axis(i, cells, X, eps, selection_angle)

            # pre-compute the X terms for each dimension
            X_neighbours = []
            for j in range(cell.dim):
                u, v = below[j], above[j]
                if not np.any(u):
                    u = None
                if not np.any(v):
                    v = None

                if u is None:
                    if v is None:
                        Xj = None
                    else:
                        Xj = 1. / (X0[j] - np.mean(X[v,j]))
                elif v is None:
                    Xj = 1. / (X0[j] - np.mean(X[u,j]))
                else:
                    Xj = np.r_[X0[j], np.mean(X[u,j]), np.mean(X[v,j])]
                if np.isscalar(Xj):
                    try:
                        Xj = Xj.tolist()
                    except AttributeError:
                        pass
                    else:
                        if isinstance(Xj, list):
                            Xj = Xj[0]

                X_neighbours.append((u, v, Xj))

            X = X_neighbours
        else:
            X = []

        if index_map is not None:
            i = index_map[i]
        cell.cache['grad1'] = (i, adjacent, X)

    if not X:
        return None

    y0, y = y[i], y[adjacent]

    # compute the gradient for each dimension separately
    grad = []
    for u, v, Xj in X: # j= dimension index
        #u, v, Xj= below, above, X term
        if u is None:
            if v is None:
                grad_j = na
            else:
                # 1./Xj = X0[j] - np.mean(X[v,j])
                grad_j = (y0 - np.mean(y[v])) * Xj
        elif v is None:
            # 1./Xj = X0[j] - np.mean(X[u,j])
            grad_j = (y0 - np.mean(y[u])) * Xj
        else:
            # Xj = np.r_[X0[j], np.mean(X[u,j]), np.mean(X[v,j])]
            grad_j = _vander(Xj, np.r_[y0, np.mean(y[u]), np.mean(y[v])])
        grad.append(grad_j)

    return np.hstack(grad)


def delta1(cells, i, X, index_map=None, eps=None, selection_angle=None):
    r"""
    Local spatial variation.

    Similar to `grad1`.
    Considering spatial coordinate :math:`x`, bin :math:`i` and its neighbour bins :math:`\\mathcal{N}_i`:

    .. math::

        \left.\Delta X_i\right|_x = \left(
            \begin{array}{ll}
                \frac{X_i - \overline{X}_{\mathcal{N}_i^-}}{x_i - \overline{x}_{\mathcal{N}_i^-}} &
                    \textrm{ or } 0 \textrm{ if } \mathcal{N}_i^- \textrm{ is } \emptyset \\
                \frac{X_i - \overline{X}_{\mathcal{N}_i^+}}{x_i - \overline{x}_{\mathcal{N}_i^+}} &
                    \textrm{ or } 0 \textrm{ if } \mathcal{N}_i^+ \textrm{ is } \emptyset \\
            \end{array}
        \right)

    Also claims cache variable *grad1* in a compatible way.

    Arguments:

        i (int):
            cell index at which the gradient is evaluated.

        X (numpy.ndarray):
            vector of a scalar measurement at every cell.

        index_map (numpy.ndarray):
            index map that converts cell indices to indices in X.

    Returns:

        numpy.ndarray:
            delta vector with as many elements as there are spatial dimensions.

    """
    cell = cells[i]
    # below, the measurement is renamed y and the coordinates are X
    y = X
    X0 = cell.center

    # cache neighbours (indices and center locations)
    if not isinstance(cell.cache, dict):
        cell.cache = {}
    try:
        i, adjacent, X = cell.cache['grad1']
    except KeyError:
        adjacent = _adjacent = cells.neighbours(i)
        if index_map is not None:
            adjacent = index_map[_adjacent]
            ok = 0 <= adjacent
            if not np.all(ok):
                adjacent, _adjacent = adjacent[ok], _adjacent[ok]
        if _adjacent.size:
            X = np.vstack([ cells[j].center for j in _adjacent ])
            below, above = neighbours_per_axis(i, cells, X, eps, selection_angle)

            # pre-compute the X terms for each dimension
            X_neighbours = []
            for j in range(cell.dim):
                u, v = below[j], above[j]
                if not np.any(u):
                    u = None
                if not np.any(v):
                    v = None

                if u is None:
                    if v is None:
                        Xj = None
                    else:
                        Xj = 1. / (X0[j] - np.mean(X[v,j]))
                elif v is None:
                    Xj = 1. / (X0[j] - np.mean(X[u,j]))
                else:
                    Xj = np.r_[X0[j], np.mean(X[u,j]), np.mean(X[v,j])]
                #if np.isscalar(Xj):
                #    try:
                #        Xj = Xj.tolist()
                #    except AttributeError:
                #        pass
                #    else:
                #        if isinstance(Xj, list):
                #            Xj = Xj[0]

                X_neighbours.append((u, v, Xj))

            X = X_neighbours
        else:
            X = []

        if index_map is not None:
            i = index_map[i]
        cell.cache['grad1'] = (i, adjacent, X)

    if not X:
        return None

    y0, y = y[i], y[adjacent]

    # compute the delta for each dimension separately
    delta = []
    for u, v, Xj in X: # j= dimension index
        #u, v, Xj= below, above, X term
        if u is None:
            if v is None:
                delta_j = np.r_[0., 0.]
            else:
                # 1./Xj = X0[j] - np.mean(X[v,j])
                delta_j = np.r_[0., (y0 - np.mean(y[v])) * Xj]
        elif v is None:
            # 1./Xj = X0[j] - np.mean(X[u,j])
            delta_j = np.r_[(y0 - np.mean(y[u])) * Xj, 0.]
        else:
            # Xj = np.r_[X0[j], np.mean(X[u,j]), np.mean(X[v,j])]
            x0, xu, xv = Xj
            delta_j = np.r_[
                (y0 - np.mean(y[u])) / (x0 - xu),
                (y0 - np.mean(y[v])) / (x0 - xv),
                ]
            #delta_j = np.mean(np.abs(delta_j))
        delta.append(delta_j)

    return np.stack(delta, axis=1) # columns must represent the space dimensions


def _vander(x, y):
    #P = poly.polyfit(x, y, 2)
    #dP = poly.polyder(P)
    #return poly.polyval(x[0], dP)
    _, b, a = poly.polyfit(x, y, 2)
    return b + 2. * a * x[0]


def setup_with_grad_arguments(setup):
    """Add :meth:`~tramway.inference.base.Distributed.grad` related arguments to inference plugin setup.

    Input argument `setup` is modified inplace.
    """
    args = setup.get('arguments', OrderedDict())
    if not ('gradient' in args or 'grad' in args):
        args['gradient'] = ('--grad', dict(help="spatial gradient implementation; any of 'grad1', 'gradn'"))
    if not ('grad_epsilon' in args or 'grad' in args):
        args['grad_epsilon'] = dict(args=('--eps', '--epsilon'), kwargs=dict(type=float, help='if defined, every spatial gradient component can recruit all of the neighbours, minus those at a projected distance less than this value'), translate=True)
    if 'grad_selection_angle' not in args:
        if 'compatibility' in args:
            compatibility_note = 'if not -c, '
        else:
            compatibility_note = ''
        args['grad_selection_angle'] = ('-a', dict(type=float, help='top angle of the selection hypercone for neighbours in the spatial gradient calculation (1= pi radians; {}default is: {})'.format(compatibility_note, default_selection_angle)))
    setup['arguments'] = args


def gradient_map(cells, feature=None, **kwargs):
    grad_kwargs = get_grad_kwargs(**kwargs)
    if feature:
        if isinstance(feature, str):
            feature = [f.strip() for f in feature.split(',')]
    else:
        raise ValueError('please define `feature`')
    grads = None
    for f in feature:
        I, X = [], []
        for i in cells:
            try:
                x = getattr(cells[i], f)
            except (SystemExit, KeyboardInterrupt):
                raise
            except:
                pass
            else:
                I.append(i)
                X.append(x)
        #I = np.array(I)
        X = np.array(X)
        reverse_index = np.full(max(cells.keys())+1, -1, dtype=int)
        reverse_index[I] = np.arange(len(I))
        index, gradX = [], []
        for i in I:
            try:
                g = cells.grad(i, X, reverse_index, **grad_kwargs)
            except (SystemExit, KeyboardInterrupt):
                raise
            except:
                g = None
            if g is not None:
                index.append(i)
                gradX.append(g)
        gradX = np.vstack(gradX)
        assert gradX.shape[1:] and gradX.shape[1] == len(cells.space_cols)
        columns = [ '{} {}'.format(f, col) for col in cells.space_cols ]
        gradX = pd.DataFrame(gradX, index=index, columns=columns)
        if grads is None:
            grads = gradX
        else:
            grads.join(gradX)
    return grads

setup = dict(
        infer='gradient_map',
        arguments=OrderedDict((
            ('feature', ('-f', dict(help='feature or comma-separated list of features which gradient is expected'))),
        )))
setup_with_grad_arguments(setup)


__all__ = ['default_selection_angle', 'get_grad_kwargs', 'neighbours_per_axis', 'grad1', 'gradn',
        'delta0', 'delta0_without_scaling', 'delta1', 'setup_with_grad_arguments', 'setup', 'gradient_map']




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .exception import GraphError, AbstractGraphError, NodeError, EdgeError, \
    MissingNodeError, MissingEdgeError, NodeAttributeError, EdgeAttributeError
from .base import Graph

__all__ = ['GraphError', 'AbstractGraphError', 'NodeError', 'EdgeError', \
    'MissingNodeError', 'MissingEdgeError', 'NodeAttributeError', 'EdgeAttributeError', 'Graph']




#!/usr/bin/env python


from tramway.helper.animation import animate_map_2d_helper


def mapviz_utility():
    import argparse
    parser = argparse.ArgumentParser(prog='mapviz',
        description='visualize animated 2D maps')
    parser.add_argument('-l', '-L', '--label', help="comma-separated list of labels")
    parser.add_argument('-f', '--feature', help="mapped feature to render")
    parser.add_argument('-r', '--frame-rate', '--fps', default=1, type=float, help="frames per second")
    parser.add_argument('-c', '--codec', help="the codec to use")
    parser.add_argument('-b', '--bit-rate', '--bitrate', type=int, help="movie bitrate")
    parser.add_argument('--dpi', type=int, help="dots per inch")
    parser.add_argument('-q', '--quiet', action='store_true', default=False, help="do not print anything")
    parser.add_argument('--write-only', action='store_true', default=False, help="do not play the movie")
    parser.add_argument('-s', '--time-step', type=float, help="time step between successive frames")
    parser.add_argument('-u', '--time-unit', type=str, default='s', help="time unit for time display")
    parser.add_argument('--colormap', help="matplotlib colormap name")
    parser.add_argument('--axis-off', '--axes-off', action='store_true', help="turn the axes off")
    parser.add_argument('--colorbar-off', action='store_true', help="turn the colorbar off")
    parser.add_argument('--bounding-box', help="bounding box as left,bottom,right,top")
    parser.add_argument('input_file', help='path to rwa file')
    parser.add_argument('output_file', nargs='?', help='path to mp4 file')
    args = parser.parse_args()

    if args.bounding_box:
        bounding_box = Bbox.from_extents(float(x) for x in args.bounding_box.split(','))
    else:
        bounding_box = None

    animate_map_2d_helper(args.input_file, args.output_file,
            label=args.label,
            feature=args.feature,
            frame_rate=args.frame_rate,
            codec=args.codec,
            bit_rate=args.bit_rate,
            play=not args.write_only,
            time_step=args.time_step,
            time_unit=args.time_unit,
            colormap=args.colormap,
            axis=not args.axis_off,
            colorbar=not args.colorbar_off,
            bounding_box=bounding_box,
            verbose=not args.quiet)


if __name__ == '__main__':
    import matplotlib
    matplotlib.use('Agg')
    mapviz_utility()




# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import numpy as np
import pandas as pd
from scipy.spatial import ConvexHull
from collections import OrderedDict


setup = {'arguments': OrderedDict((
        ('convex_hull', dict(help="'always' if the partition was made with `knn`")),
        ('trust_volume', dict(action='store_true', help="for non-Voronoi tessellations only")),
        ('volume', dict(help="volume plugin name prefix (e.g. spherical)")),
        ))}


def infer_density(cells, convex_hull='fallback', trust_volume=False, scale=False, min_volume=None,
        volume=None, **kwargs):
    """
    If ``knn`` was defined for the partition, use ``convex_hull='always'``.

    With the Voronoi-based tessellations (*random*, *kmeans*, *gwr*), do not set
    ``trust_volume`` to ``True``.
    With the other tessellations (*grid*, *hexagon*, *kdtree*), use
    ``trust_volume=True`` if ``convex_hull`` evaluates to ``True``.

    With the regular tessellations (*grid*, *hexagon*), ``scale`` can be ``True``.
    With the other tessellations, do not set ``scale`` to ``True``.
    """
    if volume is None:
        convex_hull_only = convex_hull == 'always'
        indices, densities = [], []
        for index in cells:
            cell = cells[index]
            if convex_hull_only:
                vol = np.inf
            else:
                vol = cell.volume
                if vol == 0 or np.isnan(vol):
                    vol = np.inf
            if np.isinf(vol) and convex_hull and cell.dim < len(cell):
                hull = ConvexHull(cell.r)
                vol = hull.volume
                if min_volume:
                    vol = max(min_volume, vol)
                elif trust_volume:
                    vol = max(cell.volume, vol)
            density = len(cell) / vol
            indices.append(index)
            densities.append(density)
        if scale:
            densities = np.array(densities)
            if scale is True:
                scale = cell.volume # any cell; regular tessellations only
            densities *= scale
    else:
        assert volume == 'spherical'
        plugin_name = volume+'.volume'
        from tramway.inference import plugins
        setup, module = plugins[plugin_name]
        infer_volume = getattr(module, setup['infer'])
        volume = infer_volume(cells, **kwargs)
        if volume is None:
            return None
        indices = volume.index
        location_count = np.array([[ float(len(cells[i])) ] for i in indices ])
        densities = location_count / volume.values
    return pd.DataFrame(densities, index=indices, columns=['density'])


__all__ = ['infer_density', 'setup']




# -*- coding: utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


"""
As of version 0.4.7, the tramway command is frozen and is only affected by minor fixes.
"""

import argparse
try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser
import sys
from tramway.core.plugin import *
import tramway.tessellation as tessellation
import tramway.inference as inference
import tramway.feature as feature
from .helper import *
#import tramway.core.hdf5.compat
import tramway.utils.inferencemap as inferencemap


def _parse_args(args):
    kwargs = dict(args.__dict__)
    del kwargs['func']
    input_files = []
    try:
        input_files = kwargs.pop('input')
    except KeyError:
        pass
    try:
        more_input_files = kwargs.pop('input_file')
    except KeyError:
        pass
    else:
        if more_input_files:
            if isinstance(more_input_files, list):
                input_files += more_input_files
            else:
                input_files.append(more_input_files)
    if not input_files:
        print('please specify input file(s)')
        sys.exit(1)
    for input_file in input_files:
        if not os.path.isfile(input_file):
            print("cannot find file: {}".format(input_file))
            sys.exit(1)
    try:
        comment = kwargs.pop('comment')
    except KeyError:
        pass
    else:
        if comment is not None:
            kwargs['comment'] = comment.replace('\\n', '\n')
    seed = kwargs.pop('seed', False)
    if seed is None:
        import time
        seed = int(time.time())
        print('random generator seed: {}'.format(seed))
    if seed is not False:
        seed = int(seed)
        import random, numpy
        random.seed(seed)
        numpy.random.seed(seed)
    return input_files, kwargs


def _render_cells(args):
    input_file, kwargs = _parse_args(args)
    output_file = kwargs.pop('output', None)
    fig_format = kwargs.pop('print', None)
    delaunay = kwargs.pop('delaunay', False)
    hist = kwargs.pop('histogram', '')
    if hist is None: hist = ''
    if delaunay:
        kwargs['xy_layer'] = 'delaunay'
    #del kwargs['min_location_count']
    cell_plot(input_file, output_file=output_file, fig_format=fig_format, figsize=True, \
        location_count_hist='c' in hist, cell_dist_hist='d' in hist, \
        location_dist_hist='p' in hist, **kwargs)
    sys.exit(0)


def _single_molecule(args):
    input_file, kwargs = _parse_args(args)
    output_file = kwargs.pop('output', None)
    if output_file is None:
        output_file = input_file[0]
    xyt = load_xyt(input_file, columns=list('xyt'))
    xyt['n'] = 1
    xyt = xyt[list('nxyt')]
    xyt.to_csv(output_file, sep='\t', index=False)


def _sample(method, parse_extra=None):
    def sample(args):
        input_file, kwargs = _parse_args(args)
        output_file = kwargs.pop('output', None)
        scaling = kwargs.pop('w', None)
        if scaling and not kwargs['scaling']:
            kwargs['scaling'] = 'whiten'
        min_nn = kwargs.pop('knn', None)
        max_nn = kwargs.pop('max_nn', None)
        if not (min_nn is None and max_nn is None):
            kwargs['knn'] = (min_nn, max_nn)
        if kwargs.pop('enable_time_regularization', False):
            kwargs['time_window_options'] = dict(time_dimension=True)
        elif kwargs.get('time_window_duration', None):
            kwargs['time_window_options'] = dict(time_dimension=False)
        if parse_extra:
            for extra_arg, parse_arg in parse_extra:
                kwargs[extra_arg] = parse_arg(**kwargs)
        kwargs = { kw: arg for kw, arg in kwargs.items() if arg is not None }
        tessellate(input_file, output_file=output_file, method=method, **kwargs)
        sys.exit(0)
    return sample

def _infer(mode, parse_extra=None):
    def __infer(args):
        input_file, kwargs = _parse_args(args)
        output_file = kwargs.pop('output', None)
        kwargs['mode'] = mode
        if kwargs.get('profile', False) is None:
            kwargs['profile'] = True
        if parse_extra:
            for extra_arg, parse_arg in parse_extra:
                kwargs[extra_arg] = parse_arg(**kwargs)
        kwargs = { kw: arg for kw, arg in kwargs.items() if arg is not None }
        infer(input_file[0], output_file=output_file, **kwargs)
        # kwargs: mode, localization_error, diffusivity_prior, potential_prior, jeffreys_prior
        sys.exit(0)
    return __infer

def _render_map(args):
    input_file, kwargs = _parse_args(args)
    output_file = kwargs.pop('output', None)
    fig_format = kwargs.pop('print', None)
    points = kwargs.pop('points')
    if points is not False:
        point_style = dict(alpha=.01)
        if points is not None:
            for prm in points.split(','):
                key, val = prm.split('=')
                if key in ('a', 'alpha'):
                    point_style['alpha'] = float(val)
                elif key in ('c', 'color'):
                    if val[0] == "'" and val[-1] == "'":
                        val = val[1:-1]
                    point_style['color'] = val
                elif key in ('s', 'size'):
                    point_style['size'] = int(val)
        kwargs['point_style'] = point_style
    if kwargs['delaunay'] is None:
        kwargs['delaunay'] = True
    elif kwargs['delaunay'] is False:
        del kwargs['delaunay']
    else:
        delaunay = {}
        for prm in kwargs['delaunay'].split(','):
            key, val = prm.split('=')
            if key in ('a', 'alpha'):
                delaunay['alpha'] = float(val)
            elif key in ('c', 'color'):
                if val[0] == "'" and val[-1] == "'":
                    val = val[1:-1]
                delaunay['color'] = val
        kwargs['delaunay'] = delaunay
    if kwargs['clip'] is None:
        kwargs['clip'] = 4.
    elif kwargs['clip'] == 0.:
        del kwargs['clip']
    if kwargs['title'] is None:
        kwargs.pop('title')
    map_plot0(input_file[0], output_file=output_file, fig_format=fig_format, figsize=True, show=True, **kwargs)
    sys.exit(0)

def _dump_rwa(args):
    input_files, kwargs = _parse_args(args)
    verbose = kwargs.pop('verbose', False)
    label = kwargs.pop('label')
    if label and input_files:
        labels = kwargs.pop('input_label', None)
        export = False
        if not input_files[1:]:
            cluster = kwargs.pop('cluster')
            vmesh = kwargs.pop('vmesh')
            export = cluster or vmesh
        if export:
            kwargs['cluster_file'] = cluster
            kwargs['vmesh_file'] = vmesh
            kwargs = { k: v for k, v in kwargs.items() if v is not None }
            inferencemap.export_file(input_files[0], label=labels, **kwargs)
        else:
            for input_file in input_files:
                print(' -> '.join(['in '+input_file] + [ str(l) for l in labels ]) + ':')
                analyses = load_rwa(input_file, lazy=True)
                for label in labels:
                    analyses = analyses[label]
                print('\t' + str(analyses.data).replace('\n', '\n\t'))
    else:
        for input_file in input_files:
            print('in {}:'.format(input_file))
            analyses = load_rwa(input_file, lazy=True)
            print(format_analyses(analyses, global_prefix='\t', node=lazytype, metadata=kwargs.get('metadata', False)))

def _curl(args):
    import copy
    import tramway.feature.curl
    input_file, kwargs = _parse_args(args)
    input_label = kwargs.get('input_label', kwargs.get('label', None))
    if input_file[1:]:
        raise NotImplementedError('cannot handle multiple input files')
    input_file = input_file[0]
    analyses = load_rwa(input_file, lazy=True)
    cells, maps, leaf = find_artefacts(analyses, (CellStats, Maps), input_label, return_subtree=True)
    curl = tramway.feature.curl.Curl(cells, maps)
    vector_fields = { f: vs for f, vs in curl.variables.items() if len(vs) == 2 }
    curl_name = kwargs.get('output_label', None)
    if curl_name:
        curl_name = curl_name.replace('*', '')
    else:
        curl_name = 'curl'
    distance = kwargs.get('radius', 1)
    curl_maps = copy.copy(maps)
    curl_maps.maps = None
    for f in vector_fields:
        _name = '{}<{}>_{}'.format(curl_name, f, distance)
        curl_map = curl.extract(_name, f, distance)
        if curl_maps.maps is None:
            curl_maps.maps = curl_map
        else:
            curl_maps.maps = curl_maps.maps.join(curl_map)
    if curl_maps.extra_args is None:
        curl_maps.extra_args = {}
    else:
        curl_maps.extra_args = dict(curl_maps.extra_args) # copy
    curl_maps.extra_args['radius'] = distance
    # insert `curl_maps` into `analyses`
    leaf.add(Analyses(curl_maps), label=kwargs.get('output_label', None))
    output_file = kwargs.get('output', None)
    if output_file is None:
        output_file = input_file
    save_rwa(output_file, analyses, force=output_file == input_file)

def _animate_trajectories(args):
    import matplotlib
    matplotlib.use('Agg')
    from tramway.helper.animation import animate_trajectories_2d_helper
    input_file, kwargs = _parse_args(args)
    input_label = kwargs.get('input_label', kwargs.get('label', None))
    if input_file[1:]:
        raise NotImplementedError('cannot handle multiple input files')
    input_file = input_file[0]
    output_file = kwargs.pop('output', None)
    write_only = kwargs.pop('write_only', False)
    kwargs = { kw: arg for kw, arg in kwargs.items() if arg is not None }
    if 'play' not in kwargs:
        kwargs['play'] = not write_only
    animate_trajectories_2d_helper(input_file, output_file, **kwargs)

def _animate_map(args):
    import matplotlib
    matplotlib.use('Agg')
    from tramway.helper.animation import animate_map_2d_helper
    input_file, kwargs = _parse_args(args)
    input_label = kwargs.pop('input_label', kwargs.get('label', None))
    kwargs['label'] = input_label
    if input_file[1:]:
        raise NotImplementedError('cannot handle multiple input files')
    input_file = input_file[0]
    output_file = kwargs.pop('output', None)
    write_only = kwargs.pop('write_only', False)
    kwargs = { kw: arg for kw, arg in kwargs.items() if arg is not None }
    if 'play' not in kwargs:
        kwargs['play'] = not write_only
    animate_map_2d_helper(input_file, output_file, **kwargs)


def _add_subparsers(sub, subcommand, aliases=None, help=None, title=None, target='targets'):
    if aliases:
        try:
            _parser = sub.add_parser(subcommand, aliases=aliases, help=help)
        except TypeError: # Py2
            _parser = sub.add_parser(subcommand, help=help)
    else:
        _parser = sub.add_parser(subcommand, help=help)
    if help:
        _subsub = _parser.add_subparsers(title=subcommand if title is None else title,
            description='{}; the available {} are:'.format(help, target))
    else:
        _subsub = None
    return _subsub, _parser


def main():
    verbose = '--verbose' in sys.argv
    if not verbose:
        simple = None
        for k, a in enumerate(sys.argv):
            if a[0] == '-' and a[1:] and all(c == 'v' for c in a[1:]):
                simple = not a[2:]
                break
        if simple is True:
            # note that '-v' may not be the verbose flag
            # if it appears after the first command (i.e. tessellate, infer);
            # exclude one such known case by testing the argument that comes next:
            try:
                float(sys.argv[k+1])
            except (ValueError, IndexError):
                verbose = True
        elif simple is False:
            verbose = True
    if verbose:
        tessellation.plugins.verbose = inference.plugins.verbose = True

    parser = argparse.ArgumentParser(prog='tramway',
        description='TRamWAy central command.',
        epilog='See also https://github.com/DecBayComp/TRamWAy',
        conflict_handler='resolve')
    global_arguments = [
        ('-v', '--verbose', dict(action='count', help='increase verbosity')),
        ('-i', '--input', dict(action='append', default=[],
            metavar='INPUT_FILE', help='path to input file')),
        ('-o', '--output', dict(metavar='OUTPUT_FILE', help='path to output file'))]
    for arg1, arg2, kwargs in global_arguments:
        parser.add_argument(arg1, arg2, dest=arg1[1]+'pre', **kwargs)
    sub = parser.add_subparsers(title='commands', \
        description="type '%(prog)s command --help' for additional help")


    # track
    track_name = 'track'
    track_aliases = ['preprocess']
    ksub, _ = _add_subparsers(sub, track_name, track_aliases,
            'track or preprocess location data', target='preprocesses')
    single_molecule_parser = ksub.add_parser('single')
    for short_arg, long_arg, kwargs in global_arguments:
        dest = short_arg[1:] + 'post'
        single_molecule_parser.add_argument(short_arg, long_arg, dest=dest, **kwargs)
    single_molecule_parser.set_defaults(func=_single_molecule)


    # tessellate
    tessellate_name = 'tessellate'
    tessellate_aliases = ['sample']
    load_tessellate_plugins = tessellate_name in sys.argv
    if not load_tessellate_plugins:
        load_tessellate_plugins = any(alias in sys.argv for alias in tessellate_aliases)
    tsub, _ = _add_subparsers(sub, tessellate_name, tessellate_aliases,
            'spatial and temporal binning of (trans-)locations data', target='binning methods')
    if load_tessellate_plugins:
        for method in tessellation.plugins:
            method_parser = tsub.add_parser(method)
            setup, _ = tessellation.plugins[method]
            short_args = short_options(setup.get('make_arguments', {}))
            for short_arg, long_arg, kwargs in global_arguments:
                dest = short_arg[1:] + 'post'
                if short_arg in short_args:
                    method_parser.add_argument(long_arg, dest=dest, **kwargs)
                else:
                    method_parser.add_argument(short_arg, long_arg, dest=dest, **kwargs)
            method_parser.add_argument('-l', '--label', '--output-label', help='output label')
            method_parser.add_argument('--comment', help='description message (newlines must be escaped)')
            method_parser.add_argument('-L', '--input-label', help='comma-separated list of input labels')
            method_parser.add_argument('--inplace', action='store_true', \
                help='replace the input sampling by the output one (only when --input-label is defined)')
            method_parser.add_argument('-n', '--knn', '--min-nn', '--knn-min', type=int, \
                help='minimum number of nearest neighbours; cells can overlap')
            method_parser.add_argument('-N', '--max-nn', '--knn-max', type=int, \
                help='maximum number of nearest neighbours')
            method_parser.add_argument('-r', '--radius', type=float, \
                help='selection radius for locations around the cell centers; cells can overlap')
            method_parser.add_argument('-d', '--distance', type=float, help='reference distance (default is the average translocation distance)')
            method_group = method_parser.add_mutually_exclusive_group()
            method_group.add_argument('-w', action='store_true', help='whiten the input data')
            method_group.add_argument('--scaling', choices=['whiten', 'unit'])
            method_parser.add_argument('-s', '--min-location-count', type=int, \
                help='minimum number of locations per cell; this affects the tessellation only and not directly the partition; see --knn and -ss for partition-related parameters')
            method_parser.add_argument('-ss', '--strict-min-location-count', '--min-n', type=int, \
                metavar='MIN_LOCATION_COUNT', \
                help='minimum number of locations per cell; this is enforced at partition time; cells with insufficient locations are discarded and not compensated for')
            method_parser.add_argument('--seed', nargs='?', default=False, \
                help='random generator seed (for testing purposes)')
            method_parser.add_argument('--disable-metadata', action='store_true', help="do not record additional metadata in the output file")
            translations = add_arguments(method_parser, setup.get('make_arguments', {}), name=method)
            try:
                method_parser.add_argument('input_file', nargs='*', help='path to input file(s)')
            except:
                pass
            if setup.get('window_compatible', True):
                method_parser.add_argument('--time-window-duration', type=float, help="time window duration in seconds")
                method_parser.add_argument('--time-window-shift', type=float, help="time window shift in seconds (default is equal to --time-window-duration so that there is no overlap between successive segments)")
                method_parser.add_argument('--enable-time-regularization', action="store_true", help="successive segments are connected; this enables the further inference steps to regularize across time; beware that most inference modes are NOT compatible with this option")
            method_parser.set_defaults(func=_sample(method, translations))


    # infer
    isub, _ = _add_subparsers(sub, 'infer', help='infer local parameters', target='inference modes')
    if 'infer' in sys.argv:
        for mode in inference.plugins:
            mode_parser = isub.add_parser(mode)
            setup, _ = inference.plugins[mode]
            short_args = short_options(setup.get('arguments', {}))
            for short_arg, long_arg, kwargs in global_arguments:
                dest = short_arg[1:] + 'post'
                if short_arg in short_args:
                    mode_parser.add_argument(long_arg, dest=dest, **kwargs)
                else:
                    mode_parser.add_argument(short_arg, long_arg, dest=dest, **kwargs)
            mode_parser.add_argument('-L', '--input-label', help='comma-separated list of input labels')
            mode_parser.add_argument('-l', '--output-label', help='output label')
            mode_parser.add_argument('--comment', help='description message for the output artefact (newlines must be escaped)')
            # shouldn't `inplace` be optional?
            mode_parser.add_argument('--inplace', action='store_true', \
                help='replace the input maps by the output ones')
            try:
                translations = add_arguments(mode_parser, setup['arguments'], name=mode)
            except KeyError:
                translations = None
            mode_parser.add_argument('--seed', nargs='?', default=False, help='random generator seed (for testing purposes)')
            mode_parser.add_argument('--profile', nargs='?', default=False, help='profile each individual child process if any')
            mode_parser.add_argument('--disable-metadata', action='store_true', help="do not record additional metadata in the output file")
            try:
                mode_parser.add_argument('input_file', nargs='?', help='path to input file')
            except:
                pass
            mode_parser.set_defaults(func=_infer(mode, translations))


    # dump analysis tree
    dump_parser = sub.add_parser('dump', help='print information from an .rwa file')
    dump_parser.set_defaults(func=_dump_rwa)
    for arg1, arg2, kwargs in global_arguments:
        if arg1 in ['-c', '-v', '-L']:
            dump_parser.add_argument(arg2, dest=arg1[1]+'post', **kwargs)
        else:
            dump_parser.add_argument(arg1, arg2, dest=arg1[1]+'post', **kwargs)
    dump_parser.add_argument('-c', '--cluster', metavar='FILE', help='path to cluster file')
    dump_parser.add_argument('-v', '--vmesh', metavar='FILE', help='path to vmesh file')
    dump_parser.add_argument('-L', '--label', help='comma-separated list of labels')
    dump_parser.add_argument('-m', '--metadata', action='store_true', help='print metadata as @key=val')
    dump_parser.add_argument('--eps', '--epsilon', metavar='EPSILON', type=float, help='margin for half-space gradient calculation (cluster file exports only; see also `tramway.inference.base.neighbours_per_axis`)')
    dump_parser.add_argument('--auto', action='store_true', help='infer default values for missing parameters')
    try:
        dump_parser.add_argument('input_file', nargs='?', help='path to input file')
    except:
        pass


    # extract features
    fsub, _ = _add_subparsers(sub, 'extract', help='extract features from maps', target='features')

    # extract curl
    curl_parser = fsub.add_parser('curl')
    curl_parser.set_defaults(func=_curl)
    for arg1, arg2, kwargs in global_arguments:
        if arg1 in ['-v']:
            curl_parser.add_argument(arg2, dest=arg1[1]+'post', **kwargs)
        else:
            curl_parser.add_argument(arg1, arg2, dest=arg1[1]+'post', **kwargs)
    curl_parser.add_argument('-L', '--label', '--input-label', help='comma-separated list of input labels')
    curl_parser.add_argument('-l', '--output-label', help='output label')
    curl_parser.add_argument('-r', '--radius', '-d', '--distance', type=int, default=1, help='radius in number of cells')
    try:
        curl_parser.add_argument('input_file', nargs='?', help='path to input file')
    except:
        pass


    # plot artefacts
    psub, _ = _add_subparsers(sub, 'draw', aliases=['show'], help='plot static data',
            target='types of data')

    # plot cells
    cells_parser = psub.add_parser('cells')
    cells_parser.set_defaults(func=_render_cells)
    for arg1, arg2, kwargs in global_arguments:
        cells_parser.add_argument(arg1, arg2, dest=arg1[1]+'post', **kwargs)
    cells_parser.add_argument('-L', '--input-label', help='comma-separated list of input labels')
    cells_parser.add_argument('--segment', type=int, help='time segment index (indices range from 0)')
    cells_parser.add_argument('-s', '--min-location-count', type=int, \
        help='minimum number of locations per cell')
    #cells_parser.add_argument('--min-distance', type=float, \
    #    help='minimum distance between neighbour cells (for histograms)')
    #cells_parser.add_argument('--avg-distance', type=float, \
    #    help='average distance between neighbour cells (for histograms)')
    cells_parser.add_argument('-D', '--delaunay', action='store_true', help='plot the Delaunay graph instead of the Voronoi')
    #cells_parser.add_argument('-H', '--histogram', help="plot/print additional histogram(s); any combination of 'c' (cell count histogram), 'd' (distance between neighbouring centers) and 'p' (distance between any pair of locations from distinct neighbouring centers)")
    cells_parser.add_argument('-p', '--print', choices=fig_formats, help='print figure(s) on disk instead of plotting')
    try:
        cells_parser.add_argument('input_file', nargs='?', help='path to input file')
    except:
        pass

    # plot map(s)
    _, map_parser = _add_subparsers(psub, 'map', ['maps'])
    map_parser.set_defaults(func=_render_map)
    for arg1, arg2, kwargs in global_arguments:
        map_parser.add_argument(arg1, arg2, dest=arg1[1]+'post', **kwargs)
    map_parser.add_argument('-L', '--input-label', help='comma-separated list of input labels')
    map_parser.add_argument('-f', '--feature', '--variable', help='mapped feature name')
    map_parser.add_argument('--segment', type=int, help='time segment index (indices range from 0)')
    map_parser.add_argument('-P', '--points', nargs='?', default=False, help='plot the points; options can be specified as "c=\'r\',a=0.1" (no space, no double quotes)')
    map_parser.add_argument('-D', '--delaunay', nargs='?', default=False, help='plot the Delaunay graph; options can be specified as "c=\'r\',a=0.1" (no space, no double quotes)')
    map_parser.add_argument('-cm', '--colormap', help='colormap name (see https://matplotlib.org/users/colormaps.html)')
    map_parser.add_argument('-c', '--clip', type=float, nargs='?', default=0., help='clip map by absolute values; clipping threshold can be specified as a number of interquartile distances above the median')
    map_parser.add_argument('-cb', '--colorbar', action='store_false', help='do not plot colorbar')
    map_parser.add_argument('--xlim', type=float, nargs='+', help='space-separated couple of limit values for the x axis')
    map_parser.add_argument('--ylim', type=float, nargs='+', help='space-separated couple of limit values for the y axis')
    map_parser.add_argument('--zlim', type=float, nargs='+', help='space-separated couple of limit values for the z axis (plots the landscape in 3D)')
    map_parser.add_argument('-p', '--print', choices=fig_formats, help='print figure(s) on disk instead of plotting')
    map_parser.add_argument('--dpi', type=int, help='dots per inch')
    map_parser.add_argument('--inferencemap', action='store_true', help='for field maps, size the arrows wrt cell size only')
    map_parser.add_argument('--title', help='figure title')
    map_parser.add_argument('--unit', help="add units to colorbars, in latex maths; if 'std', set diffusivity to \mu\\rm{m}^2\\rm{s}^{-1}, and potential and force amplitude to k_{\\rm{B}}T")
    try:
        map_parser.add_argument('input_file', nargs='?', help='path to input file')
    except:
        pass


    # animate artefacts
    asub, _ = _add_subparsers(sub, 'animate', aliases=['grab'], help='make a movie')

    # animate trajectories
    try:
        traj_parser = asub.add_parser('trajectories', aliases=['traj'])
    except TypeError: # Py2
        traj_parser = asub.add_parser('trajectories')
    traj_parser.set_defaults(func=_animate_trajectories)
    for arg1, arg2, kwargs in global_arguments:
        traj_parser.add_argument(arg1, arg2, dest=arg1[1]+'post', **kwargs)
    traj_parser.add_argument('-r', '--frame-rate', '--fps', type=float, help="frames per second")
    traj_parser.add_argument('-c', '--codec', help="the codec to use")
    traj_parser.add_argument('-b', '--bit-rate', '--bitrate', type=int, help="movie bitrate")
    traj_parser.add_argument('--dots-per-inch', '--dpi', type=int, help="dots per inch")
    traj_parser.add_argument('--write-only', action='store_true', default=False, help="do not play the movie")
    traj_parser.add_argument('-s', '--time-step', type=float, help="time step between successive frames")
    traj_parser.add_argument('-u', '--time-unit', type=str, default='s', help="time unit for time display")

    # animate a map
    map_parser = asub.add_parser('map')
    map_parser.set_defaults(func=_animate_map)
    for arg1, arg2, kwargs in global_arguments:
        map_parser.add_argument(arg1, arg2, dest=arg1[1]+'post', **kwargs)
    map_parser.add_argument('-l', '-L', '--label', help="comma-separated list of labels to the map")
    map_parser.add_argument('-f', '--feature', help="mapped feature to be rendered")
    map_parser.add_argument('-r', '--frame-rate', '--fps', type=float, help="frames per second")
    map_parser.add_argument('-c', '--codec', help="the codec to use")
    map_parser.add_argument('-b', '--bit-rate', '--bitrate', type=int, help="movie bitrate")
    map_parser.add_argument('--dots-per-inch', '--dpi', type=int, help="dots per inch")
    map_parser.add_argument('--write-only', action='store_true', default=False, help="do not play the movie")
    map_parser.add_argument('-s', '--time-step', type=float, help="time step between successive frames")
    map_parser.add_argument('-u', '--time-unit', type=str, default='s', help="time unit for time display")


    # parse
    args = parser.parse_args()
    args.verbose = args.vpre
    try:
        args.vpost
    except AttributeError:
        args.input = args.ipre
        args.output = args.opre
    else:
        if args.vpost:
            if args.verbose is None:
                args.verbose = args.vpost
            else:
                args.verbose += args.vpost
        args.input = args.ipre + args.ipost
        args.output = args.opre if args.opre else args.opost
        del args.vpost
        del args.ipost
        del args.opost
    del args.vpre
    del args.ipre
    del args.opre
    if args.verbose is None:
        args.verbose = False
    labels = None
    try:
        labels = args.input_label
    except AttributeError:
        try:
            labels = args.label
        except AttributeError:
            pass
    if labels:
        if labels[0] in "'\"" and labels[0] == labels[-1]:
            labels = labels[1:-1]
        if ';' in labels:
            labels = labels.split(';')
        else:
            labels = labels.split(',')
        args.input_label = []
        for label in labels:
            try:
                label = int(label)
            except (TypeError, ValueError):
                pass
            args.input_label.append(label)
    try:
        args.output_label = int(args.output_label)
    except (AttributeError, TypeError, ValueError):
        pass
    try:
        args.func(args)
    except AttributeError as e:
        if e.args and 'Namespace' in e.args[0]:
            parser.print_help()
        else:
            raise



if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-

# Copyright © 2017-2018, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from math import *
from tramway.core import *
from ..base import Voronoi, sparse_to_dict
import numpy as np
import pandas as pd
import scipy.sparse as sparse
import itertools
from collections import Counter, OrderedDict


class RegularMesh(Voronoi):
    """Regular k-D grid.

    Attributes:
        lower_bound (pandas.Series or numpy.ndarray): (scaled)
        upper_bound (pandas.Series or numpy.ndarray): (scaled)
        count_per_dim (list of ints):
        min_probability (float): (not used)
        avg_probability (float):
        max_probability (float): (not used)
        min_distance (float):
            minimum distance between adjacent cell centers;
            ignored if `avg_distance` is defined.
        avg_distance (float):
            average distance between adjacent cell centers;
            ignored if `avg_probability` is defined.

    """

    __lazy__ = Voronoi.__lazy__ + ('diagonal_adjacency',)

    def __init__(self, scaler=None, lower_bound=None, upper_bound=None, count_per_dim=None, \
        min_probability=None, max_probability=None, avg_probability=None, \
        min_distance=None, avg_distance=None, **kwargs):
        Voronoi.__init__(self, scaler)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.count_per_dim = count_per_dim
        self.min_probability = min_probability
        self.max_probability = max_probability
        self.avg_probability = avg_probability
        self.min_distance = min_distance
        self.avg_distance = avg_distance
        self._diagonal_adjacency = None

    def tessellate(self, points, **kwargs):
        points = self._preprocess(points)
        if self.lower_bound is None:
            self.lower_bound = points.min(axis=0)
        elif isinstance(points, pd.DataFrame) and not isinstance(self.lower_bound, pd.Series):
            self.lower_bound = pd.Series(self.lower_bound, index=points.columns)
        if self.upper_bound is None:
            self.upper_bound = points.max(axis=0)
        elif isinstance(points, pd.DataFrame) and not isinstance(self.upper_bound, pd.Series):
            self.upper_bound = pd.Series(self.upper_bound, index=points.columns)
        if self.count_per_dim is None:
            size = self.upper_bound - self.lower_bound
            if self.avg_probability:
                n_cells = 1.0 / self.avg_probability
                increment = exp(log(np.asarray(size).prod() / n_cells) / points.shape[1])
                if self.min_distance is not None:
                    increment = max(increment, self.min_distance)
            elif self.avg_distance:
                increment = self.avg_distance
                if self.min_probability is not None:
                    # TODO
                    pass
            else:
                raise ValueError('both `avg_probability` and `avg_distance` undefined')
            if isinstance(size, pd.Series):
                self.count_per_dim = pd.Series.round(size / increment)
            else:
                self.count_per_dim = np.round(size / increment)
        elif isinstance(points, pd.DataFrame) and not isinstance(self.count_per_dim, pd.Series):
            self.count_per_dim = pd.Series(self.count_per_dim, index=points.columns)
        _linspace = lambda _start, _stop, _nsteps: np.linspace(_start, _stop, int(_nsteps))
        if isinstance(points, pd.DataFrame):
            grid = pd.concat([self.lower_bound, self.upper_bound, self.count_per_dim + 1], axis=1).T
            self.grid = [ _linspace(*grid[col].values) for col in grid ]
        else:
            grid = np.stack((self.lower_bound, self.upper_bound, self.count_per_dim + 1), axis=0)
            self.grid = [ _linspace(col[0], col[1], col[2]) for col in grid.T ]
        cs = np.meshgrid(*[ (g[:-1] + g[1:]) / 2 for g in self.grid ], indexing='ij')
        self._cell_centers = np.column_stack([ c.flatten() for c in cs ])

    def _preprocess(self, points):
        Voronoi._preprocess(self, points) # initialize `scaler`
        return points # ... but do not scale

    def _postprocess(self):
        pass

    # cell_centers property
    @property
    def cell_centers(self):
        return self._cell_centers

    @cell_centers.setter
    def cell_centers(self, centers):
        self._cell_centers = centers

    # vertices property
    @property
    def vertices(self):
        if self._vertices is None:
            vs = np.meshgrid(*self.grid, indexing='ij')
            self._vertices = np.column_stack([ v.flatten() for v in vs ])
        return self.__returnlazy__('vertices', self._vertices)

    @vertices.setter
    def vertices(self, vertices):
        self.__lazysetter__(vertices)

    # cell_adjacency property
    @property
    def cell_adjacency(self):
        if self._cell_adjacency is None:
            cix = np.meshgrid(*[ np.arange(0, len(g) - 1) for g in self.grid ], \
                indexing='ij')
            cix = np.column_stack([ g.flatten() for g in cix ])
            c2  = np.atleast_2d(np.sum(cix * cix, axis=1))
            self._cell_adjacency = sparse.csr_matrix(\
                np.abs(c2 + c2.T - 2 * np.dot(cix, cix.T) - 1.0) < 1e-6)
        return self.__returnlazy__('cell_adjacency', self._cell_adjacency)

    @cell_adjacency.setter # copy/paste
    def cell_adjacency(self, matrix):
        self.__lazysetter__(matrix)

    # vertex_adjacency property
    @property
    def vertex_adjacency(self):
        if self._vertex_adjacency is None:
            vix = np.meshgrid(*[ np.arange(0, len(g)) for g in self.grid ], \
                indexing='ij')
            vix = np.column_stack([ g.flatten() for g in vix ])
            v2  = np.atleast_2d(np.sum(vix * vix, axis=1))
            self._vertex_adjacency = sparse.csr_matrix(\
                np.abs(v2 + v2.T - 2 * np.dot(vix, vix.T) - 1.0) < 1e-6)
        return self.__returnlazy__('vertex_adjacency', self._vertex_adjacency)

    @vertex_adjacency.setter # copy/paste
    def vertex_adjacency(self, matrix):
        self.__lazysetter__(matrix)

    # cell_vertices property
    @property
    def cell_vertices(self):
        if self._cell_vertices is None:
            cs, vs = self.cell_centers, self.vertices
            c2 = np.atleast_2d(np.sum(cs * cs, axis=1))
            v2 = np.atleast_2d(np.sum(vs * vs, axis=1))
            d2 = np.sum((cs[0] - vs[0]) ** 2)
            self._cell_vertices = sparse.dok_matrix(\
                np.abs(c2.T + v2 - 2 * np.dot(cs, vs.T) - d2) < 1e-6)
            #assert self._cell_vertices.tocsr() == dict_to_sparse(sparse_to_dict(self._cell_vertices), shape=self._cell_vertices.shape)
            self._cell_vertices = sparse_to_dict(self._cell_vertices)
        return self.__returnlazy__('cell_vertices', self._cell_vertices)

    @cell_vertices.setter # copy/paste
    def cell_vertices(self, matching):
        self.__lazysetter__(matching)

    @property
    def diagonal_adjacency(self):
        if self._diagonal_adjacency is None:
            A = self.cell_adjacency.tocsr()
            indptr, indices = np.zeros_like(A.indptr), []
            for i in range(A.shape[0]):
                I = A.indices[A.indptr[i]:A.indptr[i+1]].tolist()
                J = Counter()
                for j in I:
                    J.update(A.indices[A.indptr[j]:A.indptr[j+1]])
                I += [ j for j in J if i != j and 1 < J[j] ]
                I.sort()
                indices.append(I)
                indptr[i+1] = len(I)
            indices = np.array(list(itertools.chain(*indices)))
            indptr = np.cumsum(indptr)
            self._diagonal_adjacency = sparse.csr_matrix(
                (np.ones(indices.shape, dtype=bool), indices, indptr),
                shape=A.shape)
        return self.__returnlazy__('diagonal_adjacency', self._diagonal_adjacency)

    @diagonal_adjacency.setter
    def diagonal_adjacency(self, matrix):
        self.__lazysetter__(matrix)

    def contour(self, cell, distance=1, **kwargs):
        """
        See also :func:`tramway.feature.adjacency.contour`.
        """
        #if kwargs.get('fallback', False):
        #       warn("no fallback support", exceptions.MissingSupportWarning)
        if 'dilation_adjacency' not in kwargs:
            kwargs['dilation_adjacency'] = self.diagonal_adjacency.tocsr()
        return Voronoi.contour(self, cell, distance, **kwargs)



setup = {
    'make': RegularMesh,
    'make_arguments': OrderedDict((
        ('min_probability', ()),
        ('avg_probability', ()),
        ('max_probability', ()),
        ('min_distance', ()),
        ('avg_distance', ()),
        # avg_location_count allows to control avg_probability from the commandline
        ('avg_location_count', dict(args=('-c', '--location-count'), kwargs=dict(type=int, default=80, help='average number of locations per cell'), translate=True)),
        )),
    }

__all__ = ['RegularMesh', 'setup']




# -*- coding: utf-8 -*-

# Copyright © 2017-2018, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""This module implements the :class:`~rwa.storable.Storable` class for TRamWAy datatypes."""

#from warnings import filterwarnings
#filterwarnings('ignore', category=FutureWarning, module='h5py', lineno=36) # best in top package __init__
import sys
from rwa import *
from .store import *
from . import store

_rwa_available = True
try:
    hdf5_agnostic_modules += ['tramway.core.analyses', 'tramway.core.scaler', \
        'tramway.tessellation', 'tramway.inference', 'tramway.feature']
except NameError: # rwa is Mock in rtd
    _rwa_available = False
    __all__ = []
else:
    __all__ = ['hdf5_storable', 'hdf5_not_storable', 'lazytype', 'lazyvalue'] # from rwa
__all__ += store.__all__ + ['store'] # from .store


if sys.version_info[0] < 3:
    from .rules import *
    from . import rules
    __all__ += rules.__all__ + ['rules']


# backward compatibility trick for `CellStats` (now `Partition`)
try:
    from tramway.tessellation.base import Partition
except:
    pass
else:
    from ..lazy import Lazy
    partition_exposes = cell_stats_expose = Lazy.__slots__ + Partition.__slots__ + ('_tesselation',)
    __all__.append('cell_stats_expose') # keep it for backward compatibility
    #__all__.append('partition_exposes')
    if _rwa_available:
        hdf5_storable(default_storable(Partition, exposes=partition_exposes), agnostic=True)
        # quick'n dirty fix; [TODO] check this is legal:
        hdf5.hdf5_service.by_storable_type['tramway.tessellation.base.CellStats'] = \
                hdf5.hdf5_service.by_storable_type['tramway.tessellation.base.Partition']

try:
    from tramway.tessellation.kdtree.dichotomy import Dichotomy, ConnectedDichotomy
except:
    pass
else:
    dichotomy_exposes = ['base_edge', 'min_depth', 'max_depth', \
        'origin', 'lower_bound', 'upper_bound', \
        'min_count', 'max_count', 'subset', 'cell']
    __all__.append('dichotomy_exposes')
    if _rwa_available:
        hdf5_storable(default_storable(Dichotomy, dichotomy_exposes), agnostic=True)
    connected_dichotomy_exposes = dichotomy_exposes + ['adjacency']
    __all__.append('connected_dichotomy_exposes')
    if _rwa_available:
        hdf5_storable(default_storable(ConnectedDichotomy, connected_dichotomy_exposes), agnostic=True)


try:
    from tramway.inference.base import Maps
except:
    pass
else:

    def poke_maps(store, objname, self, container, visited=None, _stack=None, legacy=False):
        #print('poke_maps')
        sub_container = store.newContainer(objname, self, container)
        attrs = dict(self.__dict__) # dict
        attrs['maps'] = attrs.pop('_maps')
        #print(list(attrs.keys()))
        if legacy:
            # legacy format
            if callable(self.mode):
                store.poke('mode', '(callable)', sub_container, visited=visited, _stack=_stack)
                store.poke('result', self.maps, sub_container, visited=visited, _stack=_stack)
            else:
                store.poke('mode', self.mode, sub_container, visited=visited, _stack=_stack)
                store.poke(self.mode, self.maps, sub_container, visited=visited, _stack=_stack)
        else:
            #print("poke 'maps'")
            store.poke('maps', self.maps, sub_container, visited=visited, _stack=_stack)
            del attrs['maps']
        deprecated = {}
        for a in ('distributed_translocations','partition_file','tessellation_param','version'):
            deprecated[a] = attrs.pop(a, None)
        for a in attrs:
            if not (a[0] == '_' or attrs[a] is None):
                try:
                    bool(attrs[a])
                except ValueError:
                    store.poke(a, attrs[a], sub_container, visited=visited, _stack=_stack)
                    continue
                if attrs[a] or attrs[a] == 0:
                    #print("poke '{}'".format(a))
                    store.poke(a, attrs[a], sub_container, visited=visited, _stack=_stack)
        for a in deprecated:
            if deprecated[a]:
                warn('`{}` is deprecated'.format(a), DeprecationWarning)
                #print("poke '{}'".format(a))
                store.poke(a, deprecated[a], sub_container, visited=visited, _stack=_stack)

    def peek_maps(store, container, _stack=None):
        #print('peek_maps')
        read = list(Maps.__lazy__) # do not read any lazy attribute;
        # in principle no lazy attribute should be found in `container`
        try:
            mode = store.peek('mode', container, _stack=_stack)
            read.append('mode')
        except KeyError:
            mode = None
        try:
            maps = store.peek('maps', container, _stack=_stack)
            read.append('maps')
        except KeyError:
            # former standalone files
            if mode == '(callable)':
                maps = store.peek('result', container, _stack=_stack)
                read.append('result')
                mode = None
            elif mode is None:
                raise
            else:
                maps = store.peek(mode, container, _stack=_stack)
                read.append(mode)
        maps = Maps(maps, mode=mode)
        for r in container:
            if r not in read:
                setattr(maps, r, store.peek(r, container, _stack=_stack))
        return maps

    __all__ += ['poke_maps', 'peek_maps']

    if _rwa_available:
        hdf5_storable(Storable(Maps, \
            handlers=StorableHandler(poke=poke_maps, peek=peek_maps)), agnostic=True)

from ..analyses import Analyses, base
if _rwa_available:
    _analyses_storable = default_storable(Analyses, exposes=base.Analyses.__slots__)
    _analyses_storable.storable_type = 'tramway.core.analyses.Analyses'
    hdf5_storable(_analyses_storable)




# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core.analyses.base import Analyses
from tramway.plot.animation import *
import os.path

def animate_trajectories_2d_helper(input_data, *args, **kwargs):
    """
    Animate 2D trajectories.

    Arguments:

        input_data (pandas.DataFrame or tramway.core.analyses.base.Analyses or str):
            nxyt data, or analysis tree, or path to xyt file.

        columns (str or list): (comma-separated) list of column names if input data
            are to be loaded; keyworded-only; see also :func:`~tramway.core.xyt.load_xyt`.

    The other arguments are passed to :func:`~tramway.plot.animation.xyt.animate_trajectories_2d`.
    """
    from tramway.plot.animation.xyt import animate_trajectories_2d
    import pandas as pd

    columns = kwargs.pop('columns', None)

    if isinstance(input_data, pd.DataFrame):
        xyt = input_data
    elif isinstance(input_data, Analyses):
        xyt = input_data.data
    else:
        input_file = os.path.expanduser(input_data)
        if not os.path.isfile(input_file):
            raise OSError("file '{}' not found".format(input_file))
        load_kwargs = {}
        if columns is not None:
            if isinstance(columns, str):
                columns = columns.split(',')
            load_kwargs['columns'] = columns
        from tramway.core.xyt import load_xyt
        try:
            xyt = load_xyt(input_file, **load_kwargs)
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            from tramway.core.hdf5 import load_rwa
            xyt = load_rwa(input_file, lazy=True).data

    animate_trajectories_2d(xyt, *args, **kwargs)


def animate_map_2d_helper(input_data, output_file=None, label=None, feature=None, variable=None, **kwargs):
    """
    Animate 2D maps.

    Arguments:

        input_data (tramway.core.analyses.base.Analyses or str): analysis tree or path to rwa file.

        output_file (str): path to .mp4 file; if None, the video stream is dumped into
            a temporary file that is deleted after being played.

        label (str or list): path to the maps as a (comma-separated) list of labels.

        feature/variable (str): name of the mapped feature to be rendered.

    The other keyword arguments are passed to :func:`~tramway.plot.animation.map.animate_map_2d`.
    """
    from tramway.core.analyses import find_artefacts
    from tramway.tessellation.base import Partition
    from tramway.inference.base import Maps
    from tramway.plot.animation.map import animate_map_2d

    if isinstance(input_data, Analyses):
        analyses = input_data
    else:
        from tramway.core.hdf5 import load_rwa
        input_file = os.path.expanduser(input_data)
        if not os.path.isfile(input_file):
            raise OSError("file '{}' not found".format(input_file))
        analyses = load_rwa(input_file, lazy=True)

    if isinstance(label, str):
        _label = []
        for _l in label.split(','):
            try:
                _l = int(_l)
            except (TypeError, ValueError):
                pass
            _label.append(_l)
        label = _label

    cells, maps = find_artefacts(analyses, (Partition, Maps), label)

    if feature is None:
        feature = variable
    if feature is None:
        if maps.features[1:]:
            raise ValueError('multiple mapped features found: {}'.format(maps.features))
        feature = maps.features[0]
    _map = maps[feature]

    animate_map_2d(_map, cells, output_file, **kwargs)


__all__ = ['animate_trajectories_2d_helper', 'animate_map_2d_helper', 'VideoReader', 'VideoWriter', 'VideoWriterReader']





import numpy
import pandas

seed = 123456789


from tramway.core.scaler import *
class TestScaler(object):

    def example_list(self, n=5):
        return self.example_ndarray(n).tolist()

    def example_ndarray(self, n=5):
        numpy.random.seed(seed)
        return numpy.random.randn(n, 3)

    def example_structured_array(self, n=5):
        arr = self.example_ndarray(n)
        t = arr.dtype
        arr = numpy.array([ tuple(row) for row in arr ], dtype=numpy.dtype([('x', t), ('y', t), ('t', t)]))
        return arr

    def example_series(self):
        return pandas.Series(self.example_list()[0], index=list('xyt'))

    def example_dataframe(self, n=5):
        return pandas.DataFrame(self.example_ndarray(n), columns=list('xyt'))

    def _isclose(self, a, b):
        return numpy.all(numpy.isclose(numpy.asarray(a).tolist(), b, atol=1e-2, rtol=0))

    def _test_unstructured(self, scaler, unscaled_array, scaled_array):
        # test 'unscale_*' methods fail on uninitialized scalers
        try:
            scaler.unscale_point(unscaled_array, inplace=False)
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            assert True
        else:
            assert False
        # initialize
        scaler.euclidean = [0,1]
        _array = scaler.scale_point(unscaled_array, inplace=False)
        assert self._isclose(_array, scaled_array)
        # test unscale
        _array = scaler.unscale_point(scaled_array, inplace=False)
        assert self._isclose(_array, numpy.asarray(unscaled_array))
        # test inplace (last test)
        if not isinstance(unscaled_array, list):
            _array = scaler.scale_point(unscaled_array, inplace=True)
            assert _array is unscaled_array
            assert self._isclose(_array, scaled_array)

    def _test_structured(self, scaler, unscaled_array, scaled_array, *args):
        # test 'unscale_*' methods fail on uninitialized scalers
        try:
            scaler.unscale_point(unscaled_array, inplace=False)
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            assert True
        else:
            assert False
        # initialize
        scaler.euclidean = list('xy') # first 2 columns
        _array = scaler.scale_point(unscaled_array, inplace=False)
        assert self._isclose(_array, scaled_array)
        # test unscale
        _array = scaler.unscale_point(_array, inplace=False)
        assert self._isclose(_array, numpy.asarray(numpy.asarray(unscaled_array).tolist()))
        # test inplace (last test)
        _array = scaler.scale_point(unscaled_array, inplace=True)
        assert _array is unscaled_array
        assert self._isclose(_array, scaled_array)
        # test additional sample data
        for unscaled, scaled in zip(args[0::2], args[1::2]):
            _scaled = scaler.scale_point(unscaled, inplace=False)
            assert self._isclose(_scaled, scaled)

    def test_noscaling(self):
        expected_result = self.example_ndarray()
        self._test_unstructured(Scaler(), self.example_list(), expected_result)
        self._test_unstructured(Scaler(), self.example_ndarray(), expected_result)
        self._test_structured(Scaler(), self.example_structured_array(), expected_result)
        self._test_structured(Scaler(), self.example_series(), expected_result[0])
        df = self.example_dataframe()
        self._test_structured(Scaler(), df, expected_result, df.iloc[0], expected_result[0])

    def test_whitening(self):
        expected_result = self.example_ndarray()
        expected_result -= numpy.mean(expected_result, axis=0, keepdims=True)
        expected_result[:,:2] /= numpy.std(expected_result[:,:2])
        expected_result[:,-1] /= numpy.std(expected_result[:,-1])
        self._test_unstructured(whiten(), self.example_list(), expected_result)
        self._test_unstructured(whiten(), self.example_ndarray(), expected_result)
        #self._test_structured(whiten(), self.example_structured_array(), expected_result) # not supported
        #self._test_structured(whiten(), self.example_series(), expected_result[0]) # supporting Series here does not make sense
        df = self.example_dataframe()
        self._test_structured(whiten(), df, expected_result, df.iloc[0], expected_result[0])

    def test_unitrange(self):
        expected_result = self.example_ndarray()
        expected_result -= numpy.min(expected_result, axis=0, keepdims=True)
        expected_result[:,:2] /= numpy.max(expected_result[:,:2])
        expected_result[:,-1] /= numpy.max(expected_result[:,-1])
        self._test_unstructured(unitrange(), self.example_list(), expected_result)
        self._test_unstructured(unitrange(), self.example_ndarray(), expected_result)
        #self._test_structured(unitrange(), self.example_structured_array(), expected_result) # not supported
        #self._test_structured(unitrange(), self.example_series(), expected_result[0]) # supporting Series here does not make sense
        df = self.example_dataframe()
        self._test_structured(unitrange(), df, expected_result, df.iloc[0], expected_result[0])



from tramway.core.xyt import *
class TestXyt(object):

    def example_nxyt(self):
        return pandas.DataFrame([[1,.1,.1,.05],[1,.45,.45,.1],[1,.85,.65,.12],[1,1.1,.9,.15],[1,.9,1.1,.2],[1,.6,.9,.25],[1,.4,.5,.3]], columns=list('nxyt'))

    def example_bbox(self):
        return [0,0,1,1]

    def test_crop(self):
        expected_result = pandas.DataFrame([[1,.1,.1,.05,.35,.35,.05],[1,.45,.45,.1,.4,.2,.02],[1,.85,.65,.12,.25,.25,.03],[2,.6,.9,.25,-.2,-.4,.05]], columns=['n','x','y','t','dx','dy','dt'])
        tested_result = crop(self.example_nxyt(), self.example_bbox())
        assert isinstance(tested_result, pandas.DataFrame) and \
                numpy.all(tested_result.columns == expected_result.columns) and \
                tested_result.dtypes.tolist() == expected_result.dtypes.tolist() and \
                numpy.all(tested_result.index == expected_result.index) and \
                numpy.allclose(tested_result, expected_result)
        return pandas.DataFrame([[1,.1,.1,.05],[1,.45,.45,.1],[1,.85,.65,.12],[2,.6,.9,.25],[2,.4,.5,.3]], columns=list('nxyt'))
        assert crop(self.example_nxyt(), self.example_bbox(), add_deltas=False).equals(expected_result[list('nxyt')])

    def test_reindex(self):
        from io import StringIO
        df=pandas.read_csv(StringIO("""\
             n          x          y           t
78894    19782  30.030399  30.079800   93.279999
78895    19782  30.054001  30.084801   93.320000
78897    19782  30.031000  30.090000   93.400000
659881  160801  30.197300  30.087900  452.640015
659882  160801  30.129200  30.116600  452.679993
659883  160801  30.191401  30.155399  452.720001
793381  192719  30.062099  30.159700  516.520020
796888  193576  30.150600  30.155199  518.280029
842499  204282  30.147100  30.192900  541.239990
907067  219796  30.189301  30.164801  570.400024\
"""), delim_whitespace=True)
        df = reindex_trajectories(df)
        assert numpy.all(df['n'].values == numpy.r_[1,1,2,3,3,3,4,5,6,7])
        reindex_trajectories(pandas.DataFrame([], columns=list('nxyt')))


from tramway.core.analyses import *
class TestAnalyses(object):

    def example_list(self, i=1):
        return ['a', i, [], True]

    def example_dict(self, i=1):
        if i == 1:
            return dict(a=(), b=2, c='str')
        elif i == 2:
            return {'key1': 'val1', 'key2': 'val2'}

    def example_set(self):
        return set([-3, 'abc'])

    def example_tuple(self):
        return (False, {}, 1e-2)

    def example_tree(self):
        tree = Analyses('a string')
        subtree = Analyses(self.example_list())
        tree.add(subtree, label='a list', comment='heterogeneous list')
        subsubtree = Analyses(self.example_dict())
        subtree.add(subsubtree, label='a dict')
        return tree

    def example_tree_2(self):
        tree = Analyses('a string')
        tree['another list'] = self.example_list(2)
        tree['another list']['a dict'] = self.example_dict(2)
        tree['another list'].comments['a dict'] = 'basic dictionnary'
        return tree

    def example_comment(self):
        return 'example comment'

    def test_accessors(self):
        tree = self.example_tree()
        assert tree.data == 'a string'
        assert tree['a list'].data == self.example_list()
        assert tree['a list']['a dict'].data == self.example_dict()
        tree.add(self.example_set())
        tree['a list']['a dict']['a tuple'] = self.example_tuple()
        assert tree[0].data == self.example_set()
        tree['a list']['a dict'].comments['a tuple'] = self.example_comment()
        assert tree['a list']['a dict'].comments['a tuple'] == self.example_comment()
        tree['a list']['a dict']['another dict'] = {}
        tree['a list']['a dict']['another dict']['yet another dict'] = self.example_dict(2)
        art1, art2 = find_artefacts(tree, ((set, list), dict), ('a list', 'a dict'))
        assert art1 == self.example_list()
        assert art2 == self.example_dict()
        art1, art2 = find_artefacts(tree, ((set, list), dict), ('a list', 'a dict', 'another dict'))
        assert art2 == {}
        art1, art2 = find_artefacts(tree, ((set, list), dict), ('a list', 'a dict', 'another dict', 'yet another dict'))
        assert art2 == self.example_dict(2)

    def test_combine_trees(self):
        a = self.example_tree()
        b = self.example_tree_2()
        append_leaf(a, b)
        assert a.data == 'a string'
        assert len(a.labels) == 2
        print(a) # shows if test fails
        assert set(a.labels) == set(('a list', 'another list'))

        a = self.example_tree()
        a._data = None
        b._data = None
        append_leaf(a, b)
        assert a.data is None
        assert len(a.labels) == 2
        print(a) # shows if test fails
        assert set(a.labels) == set(('a list', 'another list'))




# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import AnalyzerNode
import warnings
from collections import deque
import numpy as np
import pandas as pd
from tramway.core.xyt import iter_frames, iter_trajectories
from tramway.plot.mesh import __colors__
from matplotlib import animation


class LineDealer(object):
    """
    Makes glyphs for trajectories to be drawn as lines and
    keeps track of the trajectories passed to `animate`
    so that glyphs are freed as soon as possible and
    can be assigned to new trajectories.
    """
    __slots__ = ('available_glyphs', 'active_glyphs')
    def __init__(self, ax, n=None, colors=__colors__, **kwargs):
        """
        Arguments:
        
            ax (Axes): matplotlib axes.
            
            n (int): number of lines to draw.
            
            colors (Sequence): sequence of colors.
            
        Extra keyword arguments are passed to `ax.plot`.
        """
        if n is None:
            n = len(colors)
        self.available_glyphs = deque()
        for k in range(n):
            glyph, = ax.plot([], [], '-', color=colors[k % len(colors)], **kwargs)
            self.available_glyphs.append(glyph)
        self.active_glyphs = {}
    def plot(self, trajs):
        """
        Updates the glyphs with the active trajectories.
        
        Arguments:
        
            trajs (list): list of active trajectories (dataframes).
            
        Returns:
        
            tuple: sequence of updated glyphs.
        """
        updated_glyphs = []
        previous_traj_ids = set(list(self.active_glyphs.keys()))
        current_traj_ids = set([ traj['n'].iloc[0] for traj in trajs ])
        for traj_ix in previous_traj_ids - current_traj_ids:
            glyph = self.active_glyphs.pop(traj_ix)
            glyph.set_data([], [])
            updated_glyphs.append(glyph)
            self.available_glyphs.append(glyph)
        for traj in trajs:
            if len(traj)==1:
                continue
            traj_ix = traj['n'].iloc[0]
            try:
                glyph = self.active_glyphs[traj_ix]
            except KeyError:
                if self.available_glyphs:
                    self.active_glyphs[traj_ix] = glyph = self.available_glyphs.popleft()
                else:
                    warnings.warn('not enough allocated glyphs; increase n')
                    continue
            else:
                # checks
                if True:
                    previous_data_len = len(glyph.get_xdata())
                    current_data_len = len(traj)
                    if current_data_len != previous_data_len+1:
                        print('x_prev=',glyph.get_xdata(), 'y_prev=',glyph.get_ydata())
                        print(traj)
                        warnings.warn('frames are skipped')
                #
            glyph.set_data(traj['x'].values, traj['y'].values)
            updated_glyphs.append(glyph)
        return tuple(updated_glyphs)
    def init_func(self):
        """
        To be passed as argument `init_func` to `FuncAnimation`.
        """
        return tuple(self.available_glyphs)
    def animate(self, trajs, dt):
        """
        Makes the function to be passed as second positional argument to `FuncAnimation`.
        
        Arguments:
        
            trajs (DataFrame): trajectories.
            
            dt (float): time step.
            
        Returns:
        
            callable: function that takes a list of index slices.
        """
        t0 = trajs['t'].min()
        k = int(np.round(t0/dt))
        k = [k] # wrapped in a mutable object (a list),
                # so that it is available from within _animate
        def _animate(_traj_ids):
            _t = k[0]*dt
            _trajs = []
            for _i,_j in _traj_ids:
                _traj = trajs.iloc[_i:_j]
                assert 0<_traj.shape[0]
                _until_t = _traj['t']<_t+dt/2
                if not np.any(_until_t) or _traj['t'].iloc[-1]<_t-dt/2:
                    raise RuntimeError('synchronization error\n\tt= {};\ttrajectory[\'t\']= {}'.format(_t, _traj['t'].tolist()))
                _trajs.append(_traj[_until_t])
            k[0] += 1
            return self.plot(_trajs)
        return _animate


class Mpl(AnalyzerNode):
    __slots__ = ()
    @property
    def animate_cls(self):
        return LineDealer
    def animate(self, fig, trajs=None, axes=None, xlim='auto', ylim=None, aspect='equal', **kwargs):
        """
        Arguments:

            fig (matplotlib.figure.Figure): figure.

            trajs (pandas.DataFrame): trajectories (and NOT translocations).

            axes (matplotlib.axes.Axes): figure axes.

            xlim (tuple or None): abscissa limits.

            ylim (tuple or None): ordinates limits.

            aspect (str or None): aspect ratio.

        Returns:

            matplotlib.animation.FuncAnimation: animation object.

        Extra input arguments are passed to :class:`~matplotlib.animation.FuncAnimation`
        or :class:`LineDealer` (and :meth:`~matplotlib.axes.Axes.plot`).

        Set ``xlim=None`` to let the axes as is, otherwise axis limits are adjusted to
        the data extent.

        The parent `SPTDataItem` object and the `RWAnalyzer.time` attribute
        may be accessed.
        """
        if trajs is None:
            trajs = self._parent.dataframe
        dt = self._parent.dt
        nframes = len(np.unique(np.round(trajs['t']/dt)))
        if axes is None:
            axes = fig.gca()
        #
        anim_kwargs = dict(blit=True, cache_frame_data=False, save_count=nframes, interval=1e3*dt,
                repeat=False)
        more_kwargs = dict(repeat_delay=None, fargs=None)
        more_kwargs.update(anim_kwargs)
        for kw in more_kwargs:
            try:
                arg = kwargs.pop(kw)
            except KeyError:
                pass
            else:
                if kw == 'interval':
                    if isinstance(arg, str) and arg.endswith('x'):
                        arg = float(arg[:-1]) * anim_kwargs[kw]
                    elif callable(arg):
                        arg = arg(anim_kwargs[kw])
                anim_kwargs[kw] = arg
        #
        line_kwargs = dict(linewidth=2)
        line_kwargs.update(kwargs)
        if 'lw' in kwargs:
            line_kwargs.pop('linewidth')
        #
        if isinstance(xlim, str) and xlim == 'auto':
            xlim, ylim = self._parent.bounds[['x','y']].values.T
        if xlim is not None:
            axes.set_xlim(xlim)
            axes.set_ylim(ylim)
        if aspect is not None:
            axes.set_aspect(aspect)
        #
        dt = self._parent.dt
        lines = self.animate_cls(axes, **line_kwargs)
        return animation.FuncAnimation(fig, lines.animate(trajs, dt), init_func=lines.init_func,
                frames=iter_frames(trajs, dt=dt, as_trajectory_slices=True, skip_empty_frames=False),
                **anim_kwargs)

    def plot(self, data=None, axes=None, aspect='equal', **kwargs):
        """
        Keyworded arguments are parsed based on prefixes:

        * arguments starting with '*loc_*' apply to locations drawn as points;
          `KeyError` is raised on line-styling arguments;
        * arguments starting with '*trj_*' apply to trajectories drawn as lines;
          if line style is `None` or line width is `0` or color is `None`,
          trajectories are not drawn, not even as points;
        * arguments starting with '*roi_*' apply to regions of interest drawn as
          polygons; if line style is `None` or line width is `0` or color is `None`,
          roi are not drawn, not even as points;

        All 3 types of elements are drawn with `matplotlib.axes.Axes.plot`.

        By default, trajectories are not drawn and locations are.
        If styling arguments are passed for trajectories, trajectories are drawn
        and locations are not.
        To draw both locations and trajectories as different glyphs, at least one
        styling argument should be passed for each type of elements.

        Missing elements, for example when no roi are defined, never raise an
        exception even with explicit styling arguments that would legally apply;
        these elements are simply not drawn.

        In the case multiple roi collections are available,
        argument `roi_collection` allows to control which
        collection is drawn.

        The `RWAnalyzer.spt_data` and `RWAnalyzer.roi` attributes may be accessed.

        """
        loc_kwargs, trj_kwargs, roi_kwargs, sup_kwargs = {}, {}, {}, {}

        for kw in kwargs:
            if kw.startswith('loc_'):
                loc_kwargs[kw[4:]] = kwargs[kw]
            elif kw.startswith('trj_'):
                trj_kwargs[kw[4:]] = kwargs[kw]
            elif kw.startswith('roi_'):
                roi_kwargs[kw[4:]] = kwargs[kw]
            elif kw.startswith('sup_'):
                sup_kwargs[kw[4:]] = kwargs[kw]
            else:
                self.logger.warning("ignoring argument: '{}'".format(kw))

        for kw in loc_kwargs:
            if kw.startswith('line'):
                raise KeyError("line styling argument ('{}') are not allowed for locations".format(kw))

        draw_loc = draw_trj = draw_roi = True
        draw_sup = bool(sup_kwargs)

        exclusions = ['color', 'marker', 'markersize']
        while draw_loc and exclusions:
            exclusion = exclusions.pop()
            draw_loc = bool(loc_kwargs.get(exclusion, True))

        exclusions = ['color', 'linestyle', 'linewidth']
        while draw_trj and exclusions:
            exclusion = exclusions.pop()
            draw_trj = bool(trj_kwargs.get(exclusion, True))

        exclusions = ['color', 'linestyle', 'linewidth']
        while draw_roi and exclusions:
            exclusion = exclusions.pop()
            draw_roi = bool(roi_kwargs.get(exclusion, True))

        exclusions = ['color', 'linestyle', 'linewidth']
        while draw_sup and exclusions:
            exclusion = exclusions.pop()
            draw_sup = bool(sup_kwargs.get(exclusion, True))

        if draw_loc and not trj_kwargs:
            draw_trj = False
        elif draw_trj and not loc_kwargs:
            draw_loc = False

        if axes is None:
            import matplotlib.pyplot as plt
            axes = plt.gca()

        loc_glyphs = trj_glyphs = roi_glyphs = sup_glyphs = None

        if draw_loc:
            if data is None:
                data = self._parent.dataframe
            if isinstance(data, pd.DataFrame) and not data.empty:
                kwargs = dict(linestyle='none', marker='.', markersize=4, alpha=.2, color='r')
                kwargs.update(loc_kwargs)
                loc_glyphs, = axes.plot(data['x'], data['y'], **kwargs)

        if draw_trj:
            if data is None:
                data = self._parent.dataframe
            if isinstance(data, pd.DataFrame) and not data.empty:
                nan = np.full(1, np.nan, dtype=data.dtypes['x'])
                x, y = [], []
                for trj in iter_trajectories(data[list('nxyt')], asslice=False, asarray=True, order='start'):
                    x.append(trj[:,0])
                    y.append(trj[:,1])
                    x.append(nan)
                    y.append(nan)
                x = np.concatenate(x[:-1])
                y = np.concatenate(y[:-1])
                kwargs = dict(linestyle='-', marker=None, color='r')
                kwargs.update(trj_kwargs)
                trj_glyphs, = axes.plot(x, y, **kwargs)

        if draw_sup:
            x, y = [], []
            from tramway.analyzer.roi import FullRegion
            for r in self._parent.roi.as_support_regions():
                if isinstance(r, FullRegion):
                    break # or, equivalently, continue
                _min, _max = r.bounding_box
                x.append(np.r_[_min[0],_min[0],_max[0],_max[0],_min[0],np.nan])
                y.append(np.r_[_min[1],_max[1],_max[1],_min[1],_min[1],np.nan])
            if x:
                x = np.concatenate(x)
                y = np.concatenate(y)
                kwargs = dict(linestyle='-', linewidth=1, marker=None, color='b')
                kwargs.update(sup_kwargs)
                sup_glyphs, = axes.plot(x, y, **kwargs)

        if draw_roi:
            x, y = [], []
            from tramway.analyzer.roi import FullRegion, BoundingBox
            collection = roi_kwargs.pop('collection', None)
            for r in self._parent.roi.as_individual_roi(collection=collection):
                if isinstance(r, FullRegion):
                    break # or, equivalently, continue
                assert isinstance(r, BoundingBox)
                _min, _max = r.bounding_box
                x.append(np.r_[_min[0],_min[0],_max[0],_max[0],_min[0],np.nan])
                y.append(np.r_[_min[1],_max[1],_max[1],_min[1],_min[1],np.nan])
            if x:
                x = np.concatenate(x)
                y = np.concatenate(y)
                kwargs = dict(linestyle='-', marker=None, color='g')
                kwargs.update(roi_kwargs)
                roi_glyphs, = axes.plot(x, y, **kwargs)

        if aspect is not None:
            axes.set_aspect(aspect)

        return dict(loc=loc_glyphs, trj=trj_glyphs, roi=roi_glyphs, sup=sup_glyphs)


__all__ = ['LineDealer', 'Mpl']




# -*- coding: utf-8 -*-

# Copyright © 2017-2018, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core import ChainArray
from .base import *
from warnings import warn
from math import pi, log
import numpy as np
import pandas as pd
from scipy.optimize import minimize
from collections import OrderedDict


setup = {'name':    ('unsmooth.df', 'degraded.df'),
        'provides': 'df',
        'arguments': OrderedDict((
        ('localization_error',  ('-e', dict(type=float, help='localization precision (see sigma; default is 0.03)'))),
        ('jeffreys_prior',      ('-j', dict(action='store_true', help="Jeffreys' prior"))),
        ('min_diffusivity',     dict(type=float, help='minimum diffusivity value allowed')),
        ('debug',       dict(action='store_true')))),
        'cell_sampling':    'individual'}


def df_neg_posterior(x, df, cell, sigma2, jeffreys_prior, dt_mean, min_diffusivity):
    """
    Adapted from InferenceMAP's *dfPosterior* procedure:

    .. code-block:: c++

        for (int i = 0; i < ZONES[CURRENT_ZONE].translocations; i++) {
            const double dt = ZONES[CURRENT_ZONE].dt[i];
            const double dx = ZONES[CURRENT_ZONE].dx[i];
            const double dy = ZONES[CURRENT_ZONE].dy[i];

            D_bruit = LOCALIZATION_ERROR*LOCALIZATION_ERROR/dt;

            result += - log(4.0*PI*(FxFyD[2]+D_bruit)*dt ) - pow(fabs(dx-FxFyD[2]*FxFyD[0]*dt ),2.0)/(4.0*(FxFyD[2]+D_bruit)*dt) - pow(fabs(dy-FxFyD[2]*FxFyD[1]*dt ),2.0)/(4.0*(FxFyD[2]+D_bruit)*dt);
        }

        if (JEFFREYS_PRIOR == 1) {
            result += 2.0*log(FxFyD[2]) - 2.0*log(FxFyD[2]*ZONES[CURRENT_ZONE].dtMean + LOCALIZATION_ERROR*LOCALIZATION_ERROR);
        }
        return -result;

    """
    df.update(x)
    D, F = df['D'], df['F']
    if D < min_diffusivity:
        warn(DiffusivityWarning(D, min_diffusivity))
    noise_dt = sigma2
    n = len(cell) # number of translocations
    D_dt = D * cell.dt
    denominator = 4. * (D_dt + noise_dt) # 4*(D+Dnoise)*dt
    dr_minus_drift_dt = cell.dr - np.outer(D_dt, F)
    # non-directional squared displacement
    ndsd = np.sum(dr_minus_drift_dt * dr_minus_drift_dt, axis=1)
    neg_posterior = n * log(pi) + np.sum(np.log(denominator)) + np.sum(ndsd / denominator)
    if jeffreys_prior:
        try:
            neg_posterior += 2. * (log(D * dt_mean + sigma2) - log(D))
        except ValueError as e: # math domain error
            warn(DiffusivityWarning(e))
    return neg_posterior


def infer_DF(cells, localization_error=None, jeffreys_prior=False, min_diffusivity=None, debug=False, \
        **kwargs):
    if isinstance(cells, Distributed): # multiple cells
        localization_error = cells.get_localization_error(kwargs, 0.03, True, \
                localization_error=localization_error)
        args = (localization_error, jeffreys_prior, min_diffusivity)
        index, inferred = [], []
        for i in cells:
            cell = cells[i]
            # sanity checks
            if not bool(cell):
                raise ValueError('empty cells')
            if cell.dr.shape[1] == 0:
                raise ValueError('translocation array has no column')
            if cell.dt.shape[1:]:
                raise ValueError('time deltas are structured in multiple dimensions')
            # ensure that translocations are properly oriented in time
            if not np.all(0 < cell.dt):
                warn('translocation dts are non-positive', RuntimeWarning)
                cell.dr[cell.dt < 0] *= -1.
                cell.dt[cell.dt < 0] *= -1.
            index.append(i)
            inferred.append(infer_DF(cell, *args, **kwargs))
        inferred = np.stack(inferred, axis=0)
        #D = inferred[:,0]
        #gradD = []
        #for i in index:
        #       gradD.append(cells.grad(i, D))
        #gradD = np.stack(gradD, axis=0)
        inferred = pd.DataFrame(inferred, \
            index=index, \
            columns=[ 'diffusivity' ] + \
                [ 'force ' + col for col in cells.space_cols ])
        #for j, col in enumerate(cells.space_cols):
        #       inferred['gradD '+col] = gradD[:,j]
        if debug:
            xy = np.vstack([ cells[i].center for i in index ])
            inferred = inferred.join(pd.DataFrame(xy, index=index, \
                columns=cells.space_cols))
        return inferred
    else: # single cell
        cell = cells
        dt_mean = np.mean(cell.dt)
        D_initial = np.mean(cell.dr * cell.dr) / (2. * dt_mean)
        F_initial = np.zeros(cell.dim, dtype=D_initial.dtype)
        df = ChainArray('D', D_initial, 'F', F_initial)
        if min_diffusivity is not False:
            if min_diffusivity is None:
                noise_dt = localization_error
                min_diffusivity = (1e-16 - noise_dt) / np.max(cell.dt)
            kwargs['bounds'] = [(min_diffusivity, None)] + [(None, None)] * cell.dim
        #cell.cache = None # no cache needed
        result = minimize(df_neg_posterior, df.combined, \
            args=(df, cell, localization_error, jeffreys_prior, dt_mean, min_diffusivity), \
            **kwargs)
        #df.update(result.x)
        #return (df['D'], df['F'])
        return result.x # needless to split df.combined into D and F




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from math import *
import numpy as np
from threading import Lock
import itertools


class Dichotomy(object):
    def __init__(self, points=None, min_edge=None, base_edge=None, \
        min_depth=None, max_depth=50, max_level=None, \
        min_count=1, max_count=None, \
        origin=None, center=None, lower_bound=None, upper_bound=None, \
        subset=None, cell=None):
        self.subset_lock = Lock()
        self.cell_lock = Lock()
        self.base_edge = base_edge
        self.min_depth = min_depth
        self.max_depth = max_depth
        self.origin = origin
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.min_count = min_count
        self.max_count = max_count
        try:
            if self.max_count is None:
                self.max_count = points.shape[0]
            if self.lower_bound is None:
                self.lower_bound = points.min(axis=0)
            if self.upper_bound is None:
                self.upper_bound = points.max(axis=0)
        except AttributeError:
            raise AttributeError('missing `points` input argument')
        if center is None:
            center = (self.lower_bound + self.upper_bound) / 2.0
        # define `max_edge`; `max_edge` is level-1 edge, not level 0, hence +1 in `max_depth`
        max_edge = np.max(np.maximum(self.upper_bound - center, center - self.lower_bound))
        if self.origin is None: # `center` cannot be `None`
            self.origin = center - max_edge
        if self.base_edge is None: # `max_edge` is defined
            if min_edge:
                self.max_depth = int(ceil(log(max_edge / min_edge, 2))) + 1
            self.base_edge = max_edge * 2.0 ** (1 - self.max_depth)
        else:
            if min_edge:
                raise AttributeError('`min_edge` ignored') # should be a warning instead
            self.max_depth = int(ceil(log(max_edge / self.base_edge, 2))) + 1
        if self.min_depth is None:
            if max_level is None:
                self.min_depth = 0
            else:
                self.min_depth = self.max_depth - max_level
        dim = self.origin.size
        grid = [(0, 1)] * dim
        grid = np.meshgrid(*grid, indexing='ij')
        self.unit_hypercube = np.stack([ g.flatten() for g in grid ], axis=-1)
        # `min_edge` is defined
        self.reference_length = self.base_edge * 2.0 ** np.arange(self.max_depth, -2, -1)
        if subset is None:
            if points is None:
                self.subset = {}
                self.subset_counter = 0
            else:
                self.subset = {0: (np.arange(points.shape[0]), points)}
                self.subset_counter = 1
        else:
            self.subset = subset
            if subset:
                self.subset_counter = max(subset.keys()) + 1
            else:
                self.subset_counter = 0
        if cell is None:
            self.cell = dict()
            self.cell_counter = 0
        else:
            self.cell = cell
            if cell:
                self.cell_counter = max(cell.keys()) + 1
            else:
                self.cell_counter = 0


    def split(self, points=None):
        if points is None:
            if self.subset_counter == 0:
                raise AttributeError('missing `points` input argument')
        else:
            self.subset = {0: (np.arange(points.shape[0]), points)}
            self.subset_counter = 1
            self.cell = dict()
            self.cell_counter = 0
        self._split(0, self.origin, 0)

    def _split(self, ss_ref, origin, depth):
        indices, points = self.subset[ss_ref]
        ok = depth < self.max_depth and self.min_count < points.shape[0]
        depth += 1
        ok = ok or depth < self.min_depth
        if ok:
            # split and check that every subarea has at least min_count points
            ss_refs = dict()
            lower = dict()
            for i, step in enumerate(self.unit_hypercube): # could be parallelized
                lower[i] = origin + np.float_(step) * self.reference_length[depth] # new origin
                upper = lower[i] + self.reference_length[depth] # opposite hypercube vertex
                self.subset_lock.acquire()
                r = self.subset_counter
                self.subset_counter += 1
                self.subset_lock.release()
                subset = np.all(np.logical_and(lower[i] < points, \
                    points < upper), axis=1)
                self.subset[r] = (indices[subset], points[subset])
                ss_refs[i] = r # dict for future parallelization
            if self.min_depth <= depth:
                counts = [ self.subset[r][0].size for r in ss_refs.values() ]
                ok = any([ self.max_count < c for c in counts ]) or \
                    all([ self.min_count < c for c in counts ])
            if ok:
                del self.subset[ss_ref] # save memory
            else:
                for r in ss_refs.values():
                    del self.subset[r]
        if ok:
            return self.do_split(ss_refs, lower, depth)
        else:
            depth -= 1
            self.cell_lock.acquire()
            i = self.cell_counter
            self.cell_counter += 1
            self.cell_lock.release()
            self.cell[i] = (origin, depth, ss_ref)
            return self.dont_split(i, ss_ref, origin, depth)

    def do_split(self, ss_ref, origin, depth):
        for i in ss_ref:
            self._split(ss_ref[i], origin[i], depth)

    def dont_split(self, cell_ref, ss_ref, origin, depth):
        pass



def _face_hash(v1, n1):
    '''key that identify a same face.'''
    return tuple(v1) + tuple(n1)


class ConnectedDichotomy(Dichotomy):
    def __init__(self, *args, **kwargs):
        adjacency = kwargs.pop('adjacency', None)
        Dichotomy.__init__(self, *args, **kwargs)
        self.adjacency_lock = Lock()
        dim = self.unit_hypercube.shape[1]
        eye = np.eye(dim, dtype=int)
        self.face_normal = np.vstack((eye[::-1], eye))
        self.face_vertex = np.vstack((np.zeros_like(eye), eye))
        def onface(vertex, face_vertex, face_normal):
            return np.sum(face_normal * (face_vertex - vertex), axis=1, keepdims=True) == 0
        self.exterior = np.concatenate([ onface(self.unit_hypercube, v, n) \
            for v, n in zip(self.face_vertex, self.face_normal) ], axis=1)
        if adjacency is None:
            self.adjacency = dict()
            self.edge_counter = 0
        else:
            self.adjacency = adjacency
            self.edge_counter = max(adjacency.keys()) + 1

    def merge_Faces(self, face1, face2, axis):
        #adjacency_old = dict(self.adjacency) # for debugging at the end of the method's body
        dim = self.unit_hypercube.shape[1]
        dims = np.logical_not(axis)
        cell1 = [ (c,) + self.cell[c] for c in face1 ]
        cell2 = [ (c,) + self.cell[c] for c in face2 ]
        #print((cell1, cell2))
        for c1, a1, l1, _ in cell1:
            a1 = a1[dims]
            for c2, a2, l2, _ in cell2:
                a2 = a2[dims]
                if l1 < l2:
                    origin, diagonal = a1, self.reference_length[l1]
                    center = a2 + .5 * self.reference_length[l2]
                else:
                    origin, diagonal = a2, self.reference_length[l2]
                    center = a1 + .5 * self.reference_length[l1]
                # is `center` inside the (origin, diagonal) hypercube?
                rel = center - origin
                if np.all(0 < rel) and np.sqrt(np.dot(rel, rel)) <= diagonal:
                    # connect cell1 with cell2
                    dk = 2 ** ((dim - 1) * abs(l1 - l2))
                    self.adjacency_lock.acquire()
                    k = self.edge_counter
                    self.edge_counter += dk
                    self.adjacency_lock.release()
                    self.adjacency[k] = (c1, c2)
        ## debug
        #self.plot_delaunay()

    def plot_delaunay(self):
        import matplotlib.pyplot
        plt  = matplotlib.pyplot
        x, x0 = [], np.full(2, np.nan, dtype=float)
        for k in self.adjacency:
            c1, c2 = self.adjacency[k]
            x1, l1, _ = self.cell[c1]
            x2, l2, _ = self.cell[c2]
            x1 = x1 + self.reference_length[l1 + 1]
            x2 = x2 + self.reference_length[l2 + 1]
            x.append(x0)
            x.append(x1)
            x.append(x2)
        x = np.stack(x, axis=0)
        plt.figure()
        plt.plot(x[:,0], x[:,1], 'r-')
        plt.axis('equal')
        plt.show()

    def do_split(self, ss_ref, origin, depth):
        interior_cells = dict()
        exterior_cells = dict()
        # split
        for i, step in enumerate(self.unit_hypercube):
            cells = self._split(ss_ref[i], origin[i], depth)
            # factorize interior/exterior
            for c, cs in enumerate(cells):
                if self.exterior[i, c]:
                    Cs = exterior_cells.get(c, [])
                    Cs.append(cs)
                    exterior_cells[c] = Cs
                else:
                    j = _face_hash(step + self.face_vertex[c], \
                        self.face_normal[c])
                    if j in interior_cells:
                        self.merge_Faces(interior_cells[j], cs, \
                            self.face_normal[c])
                    else:
                        interior_cells[j] = cs
        for c in exterior_cells:
            exterior_cells[c] = list(itertools.chain(*exterior_cells[c]))
        dim = self.unit_hypercube.shape[1]
        return [ exterior_cells[c] for c in range(2 * dim) ]

    def dont_split(self, cell_ref, ss_ref, origin, depth):
        dim = self.unit_hypercube.shape[1]
        return [ [cell_ref] for _ in range(2 * dim) ]





# -*- coding: utf-8 -*-

# Copyright © 2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from tramway.core.namedcolumns import *
from .base import *
from .gradient import *
import numpy as np
import pandas as pd
from collections import OrderedDict
import itertools
import warnings

_extensions = [ 'n', 'V_prior', 'V', 'zeta_total', 'zeta_spurious' ]


setup = {'name': ('snr', 'd.snr'),
    'arguments': OrderedDict((
        ('localization_error',  ('-e', dict(type=float, help='localization precision (see also sigma; default is 0.03)'))),
        ('diffusivity_prior',   ('-d', dict(type=float, help='prior on the diffusivity'))),
        ('jeffreys_prior',      ('-j', dict(action='store_true', help="Jeffreys' prior"))),
        ('min_diffusivity',     dict(type=float, help='minimum diffusivity value allowed')),
        ('max_iter',        dict(type=int, help='maximum number of iterations')))),
    'cell_sampling': 'group'}
setup_with_grad_arguments(setup)


def infer_snr(cells, **kwargs):
    '''
    Add signal-to-noise ratio related variables useful for Bayes factor calculation.

    If variable *diffusivity* is not already available, it is inferred using D mode,
    with or without spatial smoothing.
    '''
    for i in cells:
        any_cell = cells[i]
        break
    if hasattr(any_cell, 'diffusivity'):
        try:
            variables = any_cell.__slots__
        except AttributeError:
            warnings.warn('attribute `__slots__` missing in cells', RuntimeWarning)
            variables = ('diffusivity',)
        else:
            if not isinstance(variables, (tuple, list)):
                variables = (variables,)
            if 'diffusivity' not in variables:
                warnings.warn('`__slots__` does not refer to diffusivity in cells', RuntimeWarning)
                variables = list(variables)
                variables.append('diffusivity')
        index = []
        _maps = {}
        _ix = {}
        for v in variables:
            _maps[v] = []
            _ix[v] = []
        for i in cells:
            cell = cells[i]
            # sanity check similar to :func:`inference.base.smooth_infer_init`
            try:
                adjacent = cells.adjacency.indices[cells.adjacency.indptr[i]:cells.adjacency.indptr[i+1]]
                adjacent = [ c for c in adjacent if cells[c] ]
                if not adjacent:
                    continue
            except ValueError:
                continue
            #
            index.append(i)
            for v in variables:
                try:
                    _val = getattr(cell, v)
                except AttributeError:
                    _val = None
                if _val is None:
                    continue
                _ix[v].append(i)
                _maps[v].append(_val)
        index = np.array(index)
        _cols = []
        for v in variables:
            if not _maps[v]:
                continue
        _cols = list(itertools.chain(_cols))
        maps = None
        for v in variables:
            if not _maps[v]:
                continue
            _data = np.array(_maps[v])
            if not _data.shape[1:]:
                _data = _data[:,np.newaxis]
            _cols = expandcoord(v, _data.shape[1])
            _map = pd.DataFrame(_data, columns=_cols, index=_ix[v])
            if maps is None:
                maps = _map
            else:
                maps = maps.join(_map)
    else:
        import tramway.helper as helper
        if kwargs.get('diffusivity_prior', None):
            maps = helper.infer(cells, 'smooth.d', **kwargs)
        else:
            maps = helper.infer(cells, 'd', **kwargs)
    maps = add_snr_extensions(cells, maps, get_grad_kwargs(**kwargs))
    return maps


def add_snr_extensions(cells, maps=None, grad_kwargs={}, _zeta_spurious=True):
    '''
    Infer signal-to-noise ratio related variables useful for Bayes factor calculation.

    Diffusion is read from `maps` (if available) or `cells`.
    Variable/attribute `diffusivity` is expected.

    Arguments:

        cells (tramway.inference.base.Distributed): distributed cells.

        maps (pandas.DataFrame or Maps): existing maps.

        grad_kwargs (dict): keyword arguments to :met:`cells.grad`.

    Returns:

        pandas.DataFrame or Maps: `maps` aumented with the SNR variables.
    '''
    if maps is None:
        index = np.array(list(cells.keys()))
        if _zeta_spurious:
            for i in cells:
                any_cell = cells[i]
                break
            if not hasattr(any_cell, 'diffusivity'):
                raise AttributeError('missing attribute `diffusivity`; please infer diffusion first')
            D = []
    else:
        if isinstance(maps, Maps):
            _maps = maps.maps
        elif isinstance(maps, pd.DataFrame):
            _maps = maps
        else:
            raise TypeError('unsupported type for `maps`: {}'.format(type(maps)))
        index = _maps.index.values
        if _zeta_spurious:
            D = _maps['diffusivity'].values
    # compute mean displacement m and variances V and V_prior (defined at cells `index`)
    sum_pts  = lambda a: np.sum(a, axis=0, keepdims=True)
    sum_dims = lambda a: np.sum(a, axis=1, keepdims=True)
    m, n, dr, dr2 = [], [], [], []
    dts = []
    for i in index:
        cell = cells[i]
        n.append(len(cell))
        m.append(np.mean(cell.dr, axis=0, keepdims=True))
        dr.append(sum_pts(cell.dr))
        dr2.append(sum_pts(cell.dr * cell.dr))
        dts.append(cell.dt)
        if maps is None and _zeta_spurious:
            D.append(cell.diffusivity)
    n = np.array(n)
    nnz = 1 < n
    m   = np.concatenate(m, axis=0)
    n   = n[:,np.newaxis]
    dr  = np.concatenate(dr, axis=0)
    dr2 = np.concatenate(dr2, axis=0)
    V   = sum_dims(dr2 - dr * dr / n) / n #(n - 1)
    n_prior   = np.sum(n)    - n
    dr_prior  = sum_pts(dr)  - dr
    dr2_prior = sum_pts(dr2) - dr2
    V_prior   = sum_dims(dr2_prior - dr_prior * dr_prior / n_prior) / n_prior #(n_prior - 1)
    # compute zeta_total (defined at cells `index`) and zeta_spurious (defined at cells `g_index`)
    sd = np.sqrt(V)
    zeta_total = np.zeros_like(m)
    zeta_total[nnz] = m[nnz] / sd[nnz]
    if _zeta_spurious:
        if maps is None:
            D = np.array(D)
        dts = np.concatenate(dts)
        dt = np.median(dts)
        if not np.all(np.isclose(dts, dt)):
            raise ValueError('dts are not all equal')
        reverse_index = np.full(cells.adjacency.shape[0], -1, dtype=int)
        reverse_index[index] = np.arange(len(index))
        # compute diffusivity gradient g (defined at cells `g_index`)
        g_index, g = [], []
        g_defined = np.zeros(len(index), dtype=bool)
        for j, i in enumerate(index):
            gradD = cells.grad(i, D, reverse_index, **grad_kwargs)
            if gradD is not None and nnz[j]:
                # the nnz condition does not prevent the gradient to be defined
                # but such cases are excluded anyway in the calculation of zeta_spurious
                g_defined[j] = True
                g_index.append(i)
                g.append(gradD[np.newaxis,:])
        g = np.concatenate(g, axis=0)
        # zeta_spurious
        zeta_spurious = g * dt / sd[g_defined]
    # format the output
    __maps = pd.DataFrame(
        np.concatenate((n, V_prior), axis=1),#, dr, dr2
        index=index,
        columns=['n'] + \
            ['V_prior'],# + \
            #['dr '+col for col in cells.space_cols] + \
            #['dr2 '+col for col in cells.space_cols],
        )
    if maps is None:
        _maps = __maps
    else:
        columns = ['n', 'V', 'V_prior'] \
                + ['zeta_total '+col for col in cells.space_cols]
        if _zeta_spurious:
            columns += ['zeta_spurious '+col for col in cells.space_cols]
        else:
            columns.append('sd')
        _maps = _maps.drop(columns=columns, errors='ignore').join(__maps)
    _maps = _maps.join(pd.DataFrame(
        np.concatenate((V, zeta_total), axis=1)[nnz],
        index=index[nnz],
        columns=['V'] + \
            ['zeta_total '+col for col in cells.space_cols],
        ))
    if _zeta_spurious:
        _maps = _maps.join(pd.DataFrame(
            zeta_spurious,
            index=g_index,
            columns=['zeta_spurious '+col for col in cells.space_cols],
            ))
    else:
        _maps = _maps.join(pd.DataFrame(
            sd[nnz],
            index=index[nnz],
            columns=['sd'],
            ))
    if isinstance(maps, Maps):
        maps.maps = _maps
    else:
        maps = _maps
    return maps


__all__ = [ 'add_snr_extensions', 'infer_snr', 'setup' ]




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import AnalyzerNode
from tramway.core.xyt import translocations_to_trajectories


class Mpl(AnalyzerNode):
    """
    See also :class:`~tramway.analyzer.spt_data.mpl.Mpl`.
    """
    __slots__ = ()
    def animate(self, fig, trajs=None, axes=None, xlim='auto', ylim=None, aspect='equal', **kwargs):
        """
        See also :meth:`~tramway.analyzer.spt_data.mpl.Mpl.animate`.
        """
        roi = self._parent
        if trajs is None:
            trajs = translocations_to_trajectories(roi.crop())
        if xlim=='auto' and ylim is None:
            _min, _max = roi.bounding_box
            xlim, ylim = (_min[0],_max[0]), (_min[1],_max[1])
        return roi._spt_data.mpl.animate(fig, trajs, axes, xlim, ylim, aspect, **kwargs)


__all__ = ['Mpl']

